import numpy as np

from yt.frontends.moab.api import MoabHex8Dataset
from yt.testing import (
    assert_almost_equal,
    assert_equal,
    requires_file,
    units_override_check,
)
from yt.utilities.answer_testing.framework import (
    FieldValuesTest,
    data_dir_load,
    requires_ds,
)

_fields = (("moab", "flux"),)

c5 = "c5/c5.h5m"


@requires_ds(c5)
def test_cantor_5():
    np.random.seed(0x4D3D3D3)
    ds = data_dir_load(c5)
    assert_equal(str(ds), "c5")
    dso = [
        None,
        ("sphere", ("c", (0.1, "unitary"))),
        ("sphere", ("c", (0.2, "unitary"))),
    ]
    dd = ds.all_data()
    assert_almost_equal(ds.index.get_smallest_dx(), 0.00411522633744843, 10)
    assert_equal(dd[("gas", "x")].shape[0], 63 * 63 * 63)
    assert_almost_equal(
        dd[("index", "cell_volume")].in_units("code_length**3").sum(dtype="float64").d,
        1.0,
        10,
    )
    for offset_1 in [1e-9, 1e-4, 0.1]:
        for offset_2 in [1e-9, 1e-4, 0.1]:
            DLE = ds.domain_left_edge
            DRE = ds.domain_right_edge
            ray = ds.ray(DLE + offset_1 * DLE.uq, DRE - offset_2 * DRE.uq)
            assert_almost_equal(ray["dts"].sum(dtype="float64"), 1.0, 8)
    for p1 in np.random.random((5, 3)):
        for p2 in np.random.random((5, 3)):
            ray = ds.ray(p1, p2)
            assert_almost_equal(ray["dts"].sum(dtype="float64"), 1.0, 8)
    for field in _fields:
        for dobj_name in dso:
            yield FieldValuesTest(c5, field, dobj_name)


@requires_file(c5)
def test_MoabHex8Dataset():
    assert isinstance(data_dir_load(c5), MoabHex8Dataset)


@requires_file(c5)
def test_units_override():
    units_override_check(c5)



import numpy as np

from yt.data_objects.selection_objects.data_selection_objects import (
    YTSelectionContainer,
    YTSelectionContainer3D,
)
from yt.data_objects.static_output import Dataset
from yt.funcs import (
    fix_length,
    validate_3d_array,
    validate_center,
    validate_float,
    validate_object,
    validate_sequence,
)


class YTDisk(YTSelectionContainer3D):
    """
    By providing a *center*, a *normal*, a *radius* and a *height* we
    can define a cylinder of any proportion.  Only cells whose centers are
    within the cylinder will be selected.

    Parameters
    ----------
    center : array_like
        coordinate to which the normal, radius, and height all reference
    normal : array_like
        the normal vector defining the direction of lengthwise part of the
        cylinder
    radius : float
        the radius of the cylinder
    height : float
        the distance from the midplane of the cylinder to the top and
        bottom planes
    fields : array of fields, optional
        any fields to be pre-loaded in the cylinder object
    ds: ~yt.data_objects.static_output.Dataset, optional
        An optional dataset to use rather than self.ds
    field_parameters : dictionary
         A dictionary of field parameters than can be accessed by derived
         fields.
    data_source: optional
        Draw the selection from the provided data source rather than
        all data associated with the data_set

    Examples
    --------

    >>> import yt
    >>> ds = yt.load("RedshiftOutput0005")
    >>> c = [0.5, 0.5, 0.5]
    >>> disk = ds.disk(c, [1, 0, 0], (1, "kpc"), (10, "kpc"))
    """

    _type_name = "disk"
    _con_args = ("center", "_norm_vec", "radius", "height")

    def __init__(
        self,
        center,
        normal,
        radius,
        height,
        fields=None,
        ds=None,
        field_parameters=None,
        data_source=None,
    ):
        validate_center(center)
        validate_3d_array(normal)
        validate_float(radius)
        validate_float(height)
        validate_sequence(fields)
        validate_object(ds, Dataset)
        validate_object(field_parameters, dict)
        validate_object(data_source, YTSelectionContainer)
        YTSelectionContainer3D.__init__(self, center, ds, field_parameters, data_source)
        self._norm_vec = np.array(normal) / np.sqrt(np.dot(normal, normal))
        self.set_field_parameter("normal", self._norm_vec)
        self.set_field_parameter("center", self.center)
        self.height = fix_length(height, self.ds)
        self.radius = fix_length(radius, self.ds)
        self._d = -1.0 * np.dot(self._norm_vec, self.center)

    def _get_bbox(self):
        """
        Return the minimum bounding box for the disk.
        """
        # http://www.iquilezles.org/www/articles/diskbbox/diskbbox.htm
        pa = self.center + self._norm_vec * self.height
        pb = self.center - self._norm_vec * self.height
        a = pa - pb
        db = self.radius * np.sqrt(1.0 - a.d * a.d / np.dot(a.d, a.d))
        return np.minimum(pa - db, pb - db), np.maximum(pa + db, pb + db)



from functools import partial

import pytest

from yt import derived_field
from yt.fields import local_fields
from yt.testing import fake_random_ds


def test_add_field_lambda():
    ds = fake_random_ds(16)

    ds.add_field(
        ("gas", "spam"),
        lambda field, data: data["gas", "density"],
        sampling_type="cell",
    )

    # check access
    ds.all_data()["gas", "spam"]


def test_add_field_partial():
    ds = fake_random_ds(16)

    def _spam(field, data, factor):
        return factor * data["gas", "density"]

    ds.add_field(
        ("gas", "spam"),
        partial(_spam, factor=1),
        sampling_type="cell",
    )

    # check access
    ds.all_data()["gas", "spam"]


def test_add_field_arbitrary_callable():
    ds = fake_random_ds(16)

    class Spam:
        def __call__(self, field, data):
            return data["gas", "density"]

    ds.add_field(("gas", "spam"), Spam(), sampling_type="cell")

    # check access
    ds.all_data()["gas", "spam"]


def test_add_field_uncallable():
    ds = fake_random_ds(16)

    class Spam:
        pass

    with pytest.raises(TypeError, match=r"(is not a callable object)$"):
        ds.add_field(("bacon", "spam"), Spam(), sampling_type="cell")


def test_add_field_wrong_signature():
    ds = fake_random_ds(16)

    def _spam(data, field):
        return data["gas", "density"]

    with pytest.raises(
        TypeError,
        match=(
            r"Received field function <function .*> with invalid signature\. "
            r"Expected exactly 2 positional parameters \('field', 'data'\), got \('data', 'field'\)"
        ),
    ):
        ds.add_field(("bacon", "spam"), _spam, sampling_type="cell")


def test_add_field_keyword_only():
    ds = fake_random_ds(16)

    def _spam(field, *, data):
        return data["gas", "density"]

    with pytest.raises(
        TypeError,
        match=(
            r"Received field function .* with invalid signature\. "
            r"Parameters 'field' and 'data' must accept positional values \(they cannot be keyword-only\)"
        ),
    ):
        ds.add_field(
            ("bacon", "spam"),
            _spam,
            sampling_type="cell",
        )


def test_derived_field(monkeypatch):

    tmp_field_info = local_fields.LocalFieldInfoContainer(None, [], None)
    monkeypatch.setattr(local_fields, "local_fields", tmp_field_info)

    @derived_field(name="pressure", sampling_type="cell", units="dyne/cm**2")
    def _pressure(field, data):
        return (
            (data.ds.gamma - 1.0)
            * data["gas", "density"]
            * data["gas", "specific_thermal_energy"]
        )



import os
from functools import partial

import numpy as np

from yt import units
from yt._typing import KnownFieldsT
from yt.fields.field_info_container import FieldInfoContainer
from yt.frontends.ramses.io import convert_ramses_conformal_time_to_physical_age
from yt.utilities.cython_fortran_utils import FortranFile
from yt.utilities.linear_interpolators import BilinearFieldInterpolator
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.physical_constants import (
    boltzmann_constant_cgs,
    mass_hydrogen_cgs,
    mh,
    mp,
)

from .field_handlers import RTFieldFileHandler

b_units = "code_magnetic"
ra_units = "code_length / code_time**2"
rho_units = "code_density"
vel_units = "code_velocity"
pressure_units = "code_pressure"
ener_units = "code_mass * code_velocity**2"
specific_ener_units = "code_velocity**2"
ang_mom_units = "code_mass * code_velocity * code_length"
cooling_function_units = " erg * cm**3 /s"
cooling_function_prime_units = " erg * cm**3 /s/K"
flux_unit = "1 / code_length**2 / code_time"
number_density_unit = "1 / code_length**3"

known_species_masses = {
    sp: mh * v
    for sp, v in [
        ("HI", 1.0),
        ("HII", 1.0),
        ("Electron", 1.0),
        ("HeI", 4.0),
        ("HeII", 4.0),
        ("HeIII", 4.0),
        ("H2I", 2.0),
        ("H2II", 2.0),
        ("HM", 1.0),
        ("DI", 2.0),
        ("DII", 2.0),
        ("HDI", 3.0),
    ]
}

known_species_names = {
    "HI": "H_p0",
    "HII": "H_p1",
    "Electron": "El",
    "HeI": "He_p0",
    "HeII": "He_p1",
    "HeIII": "He_p2",
    "H2I": "H2_p0",
    "H2II": "H2_p1",
    "HM": "H_m1",
    "DI": "D_p0",
    "DII": "D_p1",
    "HDI": "HD_p0",
}

_cool_axes = ("lognH", "logT")  # , "logTeq")
_cool_arrs = (
    ("cooling_primordial", cooling_function_units),
    ("heating_primordial", cooling_function_units),
    ("cooling_compton", cooling_function_units),
    ("heating_compton", cooling_function_units),
    ("cooling_metal", cooling_function_units),
    ("cooling_primordial_prime", cooling_function_prime_units),
    ("heating_primordial_prime", cooling_function_prime_units),
    ("cooling_compton_prime", cooling_function_prime_units),
    ("heating_compton_prime", cooling_function_prime_units),
    ("cooling_metal_prime", cooling_function_prime_units),
    ("mu", None),
    ("abundances", None),
)
_cool_species = (
    "Electron_number_density",
    "HI_number_density",
    "HII_number_density",
    "HeI_number_density",
    "HeII_number_density",
    "HeIII_number_density",
)

_X = 0.76  # H fraction, hardcoded
_Y = 0.24  # He fraction, hardcoded


class RAMSESFieldInfo(FieldInfoContainer):
    known_other_fields: KnownFieldsT = (
        ("Density", (rho_units, ["density"], None)),
        ("x-velocity", (vel_units, ["velocity_x"], None)),
        ("y-velocity", (vel_units, ["velocity_y"], None)),
        ("z-velocity", (vel_units, ["velocity_z"], None)),
        ("Pres_IR", (pressure_units, ["pres_IR", "pressure_IR"], None)),
        ("Pressure", (pressure_units, ["pressure"], None)),
        ("Metallicity", ("", ["metallicity"], None)),
        ("HII", ("", ["H_p1_fraction"], None)),
        ("HeII", ("", ["He_p1_fraction"], None)),
        ("HeIII", ("", ["He_p2_fraction"], None)),
        ("x-acceleration", (ra_units, ["acceleration_x"], None)),
        ("y-acceleration", (ra_units, ["acceleration_y"], None)),
        ("z-acceleration", (ra_units, ["acceleration_z"], None)),
        ("Potential", (specific_ener_units, ["potential"], None)),
        ("B_x_left", (b_units, ["magnetic_field_x_left"], None)),
        ("B_x_right", (b_units, ["magnetic_field_x_right"], None)),
        ("B_y_left", (b_units, ["magnetic_field_y_left"], None)),
        ("B_y_right", (b_units, ["magnetic_field_y_right"], None)),
        ("B_z_left", (b_units, ["magnetic_field_z_left"], None)),
        ("B_z_right", (b_units, ["magnetic_field_z_right"], None)),
    )
    known_particle_fields: KnownFieldsT = (
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
        ("particle_velocity_x", (vel_units, [], None)),
        ("particle_velocity_y", (vel_units, [], None)),
        ("particle_velocity_z", (vel_units, [], None)),
        ("particle_mass", ("code_mass", [], None)),
        ("particle_identity", ("", ["particle_index"], None)),
        ("particle_refinement_level", ("", [], None)),
        ("particle_birth_time", ("code_time", ["age"], None)),
        ("conformal_birth_time", ("", [], None)),
        ("particle_metallicity", ("", [], None)),
        ("particle_family", ("", [], None)),
        ("particle_tag", ("", [], None)),
    )

    known_sink_fields: KnownFieldsT = (
        ("particle_position_x", ("code_length", [], None)),
        ("particle_position_y", ("code_length", [], None)),
        ("particle_position_z", ("code_length", [], None)),
        ("particle_velocity_x", (vel_units, [], None)),
        ("particle_velocity_y", (vel_units, [], None)),
        ("particle_velocity_z", (vel_units, [], None)),
        ("particle_mass", ("code_mass", [], None)),
        ("particle_identifier", ("", ["particle_index"], None)),
        ("particle_birth_time", ("code_time", ["age"], None)),
        ("BH_real_accretion", ("code_mass/code_time", [], None)),
        ("BH_bondi_accretion", ("code_mass/code_time", [], None)),
        ("BH_eddington_accretion", ("code_mass/code_time", [], None)),
        ("BH_esave", (ener_units, [], None)),
        ("gas_spin_x", (ang_mom_units, [], None)),
        ("gas_spin_y", (ang_mom_units, [], None)),
        ("gas_spin_z", (ang_mom_units, [], None)),
        ("BH_spin_x", ("", [], None)),
        ("BH_spin_y", ("", [], None)),
        ("BH_spin_z", ("", [], None)),
        ("BH_spin", (ang_mom_units, [], None)),
        ("BH_efficiency", ("", [], None)),
    )

    def setup_particle_fields(self, ptype):
        super().setup_particle_fields(ptype)

        def star_age(field, data):
            if data.ds.cosmological_simulation:
                conformal_age = data[ptype, "conformal_birth_time"]
                physical_age = convert_ramses_conformal_time_to_physical_age(
                    data.ds, conformal_age
                )
                return data.ds.arr(physical_age, "code_time")
            else:
                formation_time = data[ptype, "particle_birth_time"]
                return data.ds.current_time - formation_time

        self.add_field(
            (ptype, "star_age"),
            sampling_type="particle",
            function=star_age,
            units=self.ds.unit_system["time"],
        )

    def setup_fluid_fields(self):
        def _temperature(field, data):
            rv = data["gas", "pressure"] / data["gas", "density"]
            rv *= mass_hydrogen_cgs / boltzmann_constant_cgs
            return rv

        self.add_field(
            ("gas", "temperature"),
            sampling_type="cell",
            function=_temperature,
            units=self.ds.unit_system["temperature"],
        )
        self.create_cooling_fields()

        self.species_names = [
            known_species_names[fn]
            for ft, fn in self.field_list
            if fn in known_species_names
        ]

        # See if we need to load the rt fields
        rt_flag = RTFieldFileHandler.any_exist(self.ds)
        if rt_flag:  # rt run
            self.create_rt_fields()

        # Load magnetic fields
        if ("gas", "magnetic_field_x_left") in self:
            self.create_magnetic_fields()

        # Potential field
        if ("gravity", "Potential") in self:
            self.create_gravity_fields()

    def create_gravity_fields(self):
        def potential_energy(field, data):
            return data["gas", "potential"] * data["gas", "cell_mass"]

        self.add_field(
            ("gas", "potential_energy"),
            sampling_type="cell",
            function=potential_energy,
            units=self.ds.unit_system["energy"],
        )

    def create_magnetic_fields(self):
        # Calculate cell-centred magnetic fields from face-centred
        def mag_field(ax):
            def _mag_field(field, data):
                return (
                    data[("gas", f"magnetic_field_{ax}_left")]
                    + data[("gas", f"magnetic_field_{ax}_right")]
                ) / 2

            return _mag_field

        for ax in self.ds.coordinates.axis_order:
            self.add_field(
                ("gas", f"magnetic_field_{ax}"),
                sampling_type="cell",
                function=mag_field(ax),
                units=self.ds.unit_system["magnetic_field_cgs"],
            )

        def _divB(field, data):
            """Calculate magnetic field divergence"""
            out = np.zeros_like(data[("gas", "magnetic_field_x_right")])
            for ax in data.ds.coordinates.axis_order:
                out += (
                    data[("gas", f"magnetic_field_{ax}_right")]
                    - data[("gas", f"magnetic_field_{ax}_left")]
                )
            return out / data[("gas", "dx")]

        self.add_field(
            ("gas", "magnetic_field_divergence"),
            sampling_type="cell",
            function=_divB,
            units=self.ds.unit_system["magnetic_field_cgs"]
            / self.ds.unit_system["length"],
        )

    def create_rt_fields(self):
        self.ds.fluid_types += ("rt",)
        p = RTFieldFileHandler.get_rt_parameters(self.ds).copy()
        p.update(self.ds.parameters)
        ngroups = p["nGroups"]
        rt_c = p["rt_c_frac"] * units.c / (p["unit_l"] / p["unit_t"])
        dens_conv = (p["unit_np"] / rt_c).value / units.cm**3

        ########################################
        # Adding the fields in the hydro_* files
        def _temp_IR(field, data):
            rv = data["gas", "pres_IR"] / data["gas", "density"]
            rv *= mass_hydrogen_cgs / boltzmann_constant_cgs
            return rv

        self.add_field(
            ("gas", "temp_IR"),
            sampling_type="cell",
            function=_temp_IR,
            units=self.ds.unit_system["temperature"],
        )

        def _species_density(field, data, species: str):
            return data["gas", f"{species}_fraction"] * data["gas", "density"]

        def _species_mass(field, data, species: str):
            return data["gas", f"{species}_density"] * data["index", "cell_volume"]

        for species in ["H_p1", "He_p1", "He_p2"]:
            self.add_field(
                ("gas", species + "_density"),
                sampling_type="cell",
                function=partial(_species_density, species=species),
                units=self.ds.unit_system["density"],
            )

            self.add_field(
                ("gas", species + "_mass"),
                sampling_type="cell",
                function=partial(_species_mass, species=species),
                units=self.ds.unit_system["mass"],
            )

        ########################################
        # Adding the fields in the rt_ files
        def gen_pdens(igroup):
            def _photon_density(field, data):
                rv = data["ramses-rt", f"Photon_density_{igroup + 1}"] * dens_conv
                return rv

            return _photon_density

        for igroup in range(ngroups):
            self.add_field(
                ("rt", f"photon_density_{igroup + 1}"),
                sampling_type="cell",
                function=gen_pdens(igroup),
                units=self.ds.unit_system["number_density"],
            )

        flux_conv = p["unit_pf"] / units.cm**2 / units.s

        def gen_flux(key, igroup):
            def _photon_flux(field, data):
                rv = data["ramses-rt", f"Photon_flux_{key}_{igroup + 1}"] * flux_conv
                return rv

            return _photon_flux

        flux_unit = (
            1 / self.ds.unit_system["time"] / self.ds.unit_system["length"] ** 2
        ).units
        for key in "xyz":
            for igroup in range(ngroups):
                self.add_field(
                    ("rt", f"photon_flux_{key}_{igroup + 1}"),
                    sampling_type="cell",
                    function=gen_flux(key, igroup),
                    units=flux_unit,
                )

    def create_cooling_fields(self):
        num = os.path.basename(self.ds.parameter_filename).split(".")[0].split("_")[1]
        filename = "%s/cooling_%05i.out" % (
            os.path.dirname(self.ds.parameter_filename),
            int(num),
        )

        if not os.path.exists(filename):
            mylog.warning("This output has no cooling fields")
            return

        # Function to create the cooling fields
        def _create_field(name, interp_object, unit):
            def _func(field, data):
                shape = data[("gas", "temperature")].shape
                d = {
                    "lognH": np.log10(_X * data[("gas", "density")] / mh).ravel(),
                    "logT": np.log10(data[("gas", "temperature")]).ravel(),
                }
                rv = interp_object(d).reshape(shape)
                if name[-1] != "mu":
                    rv = 10 ** interp_object(d).reshape(shape)
                cool = data.ds.arr(rv, unit)
                if "metal" in name[-1].split("_"):
                    cool = (
                        cool * data[("gas", "metallicity")] / 0.02
                    )  # Ramses uses Zsolar=0.02
                elif "compton" in name[-1].split("_"):
                    cool = data.ds.arr(rv, unit + "/cm**3")
                    cool = (
                        cool / data[("gas", "number_density")]
                    )  # Compton cooling/heating is written to file in erg/s
                return cool

            self.add_field(name=name, sampling_type="cell", function=_func, units=unit)

        # Load cooling files
        avals = {}
        tvals = {}
        with FortranFile(filename) as fd:
            n1, n2 = fd.read_vector("i")
            for ax in _cool_axes:
                avals[ax] = fd.read_vector("d")
            for i, (tname, unit) in enumerate(_cool_arrs):
                var = fd.read_vector("d")
                if var.size == n1 and i == 0:
                    # If this case occurs, the cooling files were produced pre-2010 in
                    # a format that is no longer supported
                    mylog.warning(
                        "This cooling file format is no longer supported. "
                        "Cooling field loading skipped."
                    )
                    return
                if var.size == n1 * n2:
                    tvals[tname] = dict(
                        data=var.reshape((n1, n2), order="F"), unit=unit
                    )
                else:
                    var = var.reshape((n1, n2, var.size // (n1 * n2)), order="F")
                    for i in range(var.shape[-1]):
                        tvals[_cool_species[i]] = dict(
                            data=var[:, :, i], unit="1/cm**3"
                        )

        # Add the mu field first, as it is needed for the number density
        interp = BilinearFieldInterpolator(
            tvals["mu"]["data"],
            (avals["lognH"], avals["logT"]),
            ["lognH", "logT"],
            truncate=True,
        )
        _create_field(("gas", "mu"), interp, tvals["mu"]["unit"])

        # Add the number density field, based on mu
        def _number_density(field, data):
            return data[("gas", "density")] / mp / data[("gas", "mu")]

        self.add_field(
            name=("gas", "number_density"),
            sampling_type="cell",
            function=_number_density,
            units=number_density_unit,
        )

        # Add the cooling and heating fields, which need the number density field
        for key in tvals:
            if key != "mu":
                interp = BilinearFieldInterpolator(
                    tvals[key]["data"],
                    (avals["lognH"], avals["logT"]),
                    ["lognH", "logT"],
                    truncate=True,
                )
                _create_field(("gas", key), interp, tvals[key]["unit"])

        # Add total cooling and heating fields
        def _all_cool(field, data):
            return (
                data[("gas", "cooling_primordial")]
                + data[("gas", "cooling_metal")]
                + data[("gas", "cooling_compton")]
            )

        def _all_heat(field, data):
            return (
                data[("gas", "heating_primordial")] + data[("gas", "heating_compton")]
            )

        self.add_field(
            name=("gas", "cooling_total"),
            sampling_type="cell",
            function=_all_cool,
            units=cooling_function_units,
        )
        self.add_field(
            name=("gas", "heating_total"),
            sampling_type="cell",
            function=_all_heat,
            units=cooling_function_units,
        )

        # Add net cooling fields
        def _net_cool(field, data):
            return data[("gas", "cooling_total")] - data[("gas", "heating_total")]

        self.add_field(
            name=("gas", "cooling_net"),
            sampling_type="cell",
            function=_net_cool,
            units=cooling_function_units,
        )



import re

import pytest

from yt.testing import fake_amr_ds
from yt.visualization.plot_window import SlicePlot


@pytest.mark.parametrize(
    ("origin", "msg"),
    [
        (
            "ONE",
            (
                "Invalid origin argument. "
                "Single element specification must be 'window', 'domain', or 'native'. "
                "Got 'ONE'"
            ),
        ),
        (
            "ONE-TWO",
            (
                "Invalid origin argument. "
                "Using 2 elements:\n"
                " - the first one must be 'left', 'right', 'lower', 'upper' or 'center'\n"
                " - the second one must be 'window', 'domain', or 'native'\n"
                "Got 'ONE-TWO'"
            ),
        ),
        (
            "ONE-window",
            (
                "Invalid origin argument. "
                "Using 2 elements:\n"
                " - the first one must be 'left', 'right', 'lower', 'upper' or 'center'\n"
                "Got 'ONE-window'"
            ),
        ),
        (
            "left-TWO",
            (
                "Invalid origin argument. "
                "Using 2 elements:\n"
                " - the second one must be 'window', 'domain', or 'native'\n"
                "Got 'left-TWO'"
            ),
        ),
        (
            "ONE-TWO-THREE",
            (
                "Invalid origin argument. "
                "Using 3 elements:\n"
                " - the first one must be 'lower', 'upper' or 'center' or a distance\n"
                " - the second one must be 'left', 'right', 'center' or a distance\n"
                " - the third one must be 'window', 'domain', or 'native'\n"
                "Got 'ONE-TWO-THREE'"
            ),
        ),
        (
            "ONE-TWO-window",
            (
                "Invalid origin argument. "
                "Using 3 elements:\n"
                " - the first one must be 'lower', 'upper' or 'center' or a distance\n"
                " - the second one must be 'left', 'right', 'center' or a distance\n"
                "Got 'ONE-TWO-window'"
            ),
        ),
        (
            "ONE-left-window",
            (
                "Invalid origin argument. "
                "Using 3 elements:\n"
                " - the first one must be 'lower', 'upper' or 'center' or a distance\n"
                "Got 'ONE-left-window'"
            ),
        ),
        (
            "ONE-left-THREE",
            (
                "Invalid origin argument. "
                "Using 3 elements:\n"
                " - the first one must be 'lower', 'upper' or 'center' or a distance\n"
                " - the third one must be 'window', 'domain', or 'native'\n"
                "Got 'ONE-left-THREE'"
            ),
        ),
        (
            "lower-left-THREE",
            (
                "Invalid origin argument. "
                "Using 3 elements:\n"
                " - the third one must be 'window', 'domain', or 'native'\n"
                "Got 'lower-left-THREE'"
            ),
        ),
        (
            ("ONE", "TWO", "THREE"),
            (
                "Invalid origin argument. "
                "Using 3 elements:\n"
                " - the first one must be 'lower', 'upper' or 'center' or a distance\n"
                " - the second one must be 'left', 'right', 'center' or a distance\n"
                " - the third one must be 'window', 'domain', or 'native'\n"
                "Got ('ONE', 'TWO', 'THREE')"
            ),
        ),
        (
            ("ONE", "TWO", (1, 1, 3)),
            (
                "Invalid origin argument. "
                "Using 3 elements:\n"
                " - the first one must be 'lower', 'upper' or 'center' or a distance\n"
                " - the second one must be 'left', 'right', 'center' or a distance\n"
                " - the third one must be 'window', 'domain', or 'native'\n"
                "Got ('ONE', 'TWO', (1, 1, 3))"
            ),
        ),
        (
            "ONE-TWO-THREE-FOUR",
            (
                "Invalid origin argument with too many elements; "
                "expected 1, 2 or 3 elements, got 'ONE-TWO-THREE-FOUR', counting 4 elements. "
                "Use '-' as a separator for string arguments."
            ),
        ),
    ],
)
def test_invalidate_origin_value(origin, msg):
    ds = fake_amr_ds(fields=[("gas", "density")], units=["g*cm**-3"])
    with pytest.raises(ValueError, match=re.escape(msg)):
        SlicePlot(ds, "z", ("gas", "density"), origin=origin)


@pytest.mark.parametrize(
    "origin",
    # don't attempt to match exactly a TypeError message because it should be
    # emitted by unyt, not yt
    [
        ((50, 50, 50), "TWO", "THREE"),
        ("ONE", (50, 50, 50), "THREE"),
    ],
)
def test_invalidate_origin_type(origin):
    ds = fake_amr_ds(fields=[("gas", "density")], units=["g*cm**-3"])
    with pytest.raises(TypeError):
        SlicePlot(ds, "z", ("gas", "density"), origin=origin)



import numpy as np

from yt.loaders import load_uniform_grid
from yt.testing import assert_equal, fake_random_ds


def test_exclude_above():
    the_ds = fake_random_ds(ndims=3)
    all_data = the_ds.all_data()
    new_ds = all_data.exclude_above(("gas", "density"), 1)
    assert_equal(new_ds[("gas", "density")], all_data[("gas", "density")])
    new_ds = all_data.exclude_above(("gas", "density"), 1e6, "g/m**3")
    assert_equal(new_ds[("gas", "density")], all_data[("gas", "density")])
    new_ds = all_data.exclude_above(("gas", "density"), 0)
    assert_equal(new_ds[("gas", "density")], [])


def test_exclude_below():
    the_ds = fake_random_ds(ndims=3)
    all_data = the_ds.all_data()
    new_ds = all_data.exclude_below(("gas", "density"), 1)
    assert_equal(new_ds[("gas", "density")], [])
    new_ds = all_data.exclude_below(("gas", "density"), 1e6, "g/m**3")
    assert_equal(new_ds[("gas", "density")], [])
    new_ds = all_data.exclude_below(("gas", "density"), 0)
    assert_equal(new_ds[("gas", "density")], all_data[("gas", "density")])


def test_exclude_nan():
    test_array = np.nan * np.ones((10, 10, 10))
    test_array[1, 1, :] = 1
    data = dict(density=test_array)
    ds = load_uniform_grid(data, test_array.shape, length_unit="cm", nprocs=1)
    ad = ds.all_data()
    no_nan_ds = ad.exclude_nan(("gas", "density"))
    assert_equal(no_nan_ds[("gas", "density")], np.array(np.ones(10)))


def test_equal():
    test_array = np.ones((10, 10, 10))
    test_array[1, 1, :] = 2.0
    test_array[2, 1, :] = 3.0
    data = dict(density=test_array)
    ds = load_uniform_grid(data, test_array.shape, length_unit="cm", nprocs=1)
    ad = ds.all_data()
    no_ones = ad.exclude_equal(("gas", "density"), 1.0)
    assert np.all(no_ones[("gas", "density")] != 1.0)
    only_ones = ad.include_equal(("gas", "density"), 1.0)
    assert np.all(only_ones[("gas", "density")] == 1.0)


def test_inside_outside():
    test_array = np.ones((10, 10, 10))
    test_array[1, 1, :] = 2.0
    test_array[2, 1, :] = 3.0
    data = dict(density=test_array)
    ds = load_uniform_grid(data, test_array.shape, length_unit="cm", nprocs=1)
    ad = ds.all_data()

    only_ones_and_twos = ad.include_inside(("gas", "density"), 0.9, 2.1)
    assert np.all(only_ones_and_twos[("gas", "density")] != 3.0)
    assert len(only_ones_and_twos[("gas", "density")]) == 990

    only_ones_and_twos = ad.exclude_outside(("gas", "density"), 0.9, 2.1)
    assert len(only_ones_and_twos[("gas", "density")]) == 990
    assert np.all(only_ones_and_twos[("gas", "density")] != 3.0)

    only_threes = ad.include_outside(("gas", "density"), 0.9, 2.1)
    assert np.all(only_threes[("gas", "density")] == 3)
    assert len(only_threes[("gas", "density")]) == 10

    only_threes = ad.include_outside(("gas", "density"), 0.9, 2.1)
    assert np.all(only_threes[("gas", "density")] == 3)
    assert len(only_threes[("gas", "density")]) == 10

    # Repeat, but convert units to g/m**3
    only_ones_and_twos = ad.include_inside(("gas", "density"), 0.9e6, 2.1e6, "g/m**3")
    assert np.all(only_ones_and_twos[("gas", "density")] != 3.0)
    assert len(only_ones_and_twos[("gas", "density")]) == 990

    only_ones_and_twos = ad.exclude_outside(("gas", "density"), 0.9e6, 2.1e6, "g/m**3")
    assert len(only_ones_and_twos[("gas", "density")]) == 990
    assert np.all(only_ones_and_twos[("gas", "density")] != 3.0)

    only_threes = ad.include_outside(("gas", "density"), 0.9e6, 2.1e6, "g/m**3")
    assert np.all(only_threes[("gas", "density")] == 3)
    assert len(only_threes[("gas", "density")]) == 10

    only_threes = ad.include_outside(("gas", "density"), 0.9e6, 2.1e6, "g/m**3")
    assert np.all(only_threes[("gas", "density")] == 3)
    assert len(only_threes[("gas", "density")]) == 10



# distutils: language = c++
# distutils: extra_compile_args = CPP14_FLAG
"""This is a wrapper around the C++ class to efficiently cast rays into an octree.
It relies on the seminal paper by  J. Revelles,, C.Ureña and M.Lastra.
"""


cimport numpy as np

import numpy as np

cimport cython
from libcpp.vector cimport vector

from cython.parallel import parallel, prange

from libc.stdlib cimport free, malloc

from .grid_traversal cimport sampler_function
from .image_samplers cimport ImageAccumulator, ImageSampler
from .partitioned_grid cimport PartitionedGrid
from .volume_container cimport VolumeContainer

DEF Nch = 4


cdef class _OctreeRayTracing:
    def __init__(self, np.ndarray LE, np.ndarray RE, int depth):
        cdef double* LE_ptr = <double *>LE.data
        cdef double* RE_ptr = <double *>RE.data
        self.oct = new Octree[int](depth, LE_ptr, RE_ptr)
        self.depth = depth

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def add_nodes(self, int[:, :] ipos_view, int[:] lvl_view, int[:] key):
        cdef int i
        cdef int ii[3]

        for i in range(len(key)):
            ii[0] = ipos_view[i, 0]
            ii[1] = ipos_view[i, 1]
            ii[2] = ipos_view[i, 2]
            self.oct.insert_node_no_ret(ii, lvl_view[i], <int> key[i])

    def __dealloc__(self):
        del self.oct



"""
    Chimera-specific IO functions



"""

import numpy as np
import unyt as un

from yt.frontends.chimera.data_structures import _find_files
from yt.utilities.io_handler import BaseIOHandler
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _h5py as h5py


class ChimeraIOHandler(BaseIOHandler):
    _particle_reader = False
    _dataset_type = "chimera"

    def __init__(self, ds):
        super().__init__(ds)
        self._handle = ds._handle
        self.filename = ds.filename

    def _read_particle_coords(self, chunks, ptf):
        # This needs to *yield* a series of tuples of (ptype, (x, y, z)).
        # chunks is a list of chunks, and ptf is a dict where the keys are
        # ptypes and the values are lists of fields.
        pass

    def _read_particle_fields(self, chunks, ptf, selector):
        # This gets called after the arrays have been allocated.  It needs to
        # yield ((ptype, field), data) where data is the masked results of
        # reading ptype, field and applying the selector to the data read in.
        # Selector objects have a .select_points(x,y,z) that returns a mask, so
        # you need to do your masking here.
        pass

    def _read_fluid_selection(self, chunks, selector, fields, size):
        rv = {}
        nodal_fields = []
        for field in fields:
            finfo = self.ds.field_info[field]
            nodal_flag = finfo.nodal_flag
            if np.any(nodal_flag):
                num_nodes = 2 ** sum(nodal_flag)
                rv[field] = np.empty((size, num_nodes), dtype="=f8")
                nodal_fields.append(field)
            else:
                rv[field] = np.empty(size, dtype="=f8")
        ind = 0
        for field, mesh, data in self.io_iter(chunks, fields):
            if data is None:
                continue
            else:
                ind += mesh.select(selector, data.flatten(), rv[field], ind)  # caches
        return rv

    def io_iter(self, chunks, fields):
        for n, chunk in enumerate(chunks):
            file = _find_files(self.filename)
            with h5py.File(file[n], "r") as f:
                # Generates mask according to the "ongrid_mask" variable
                m = int(file[n][-5:-3]) - 1
                k = f["fluid"]["entropy"].shape[0]
                mask_0 = f["mesh"]["ongrid_mask"][k * m : k * (m + 1), :]

                if f["mesh"]["array_dimensions"][2] > 1:
                    nrd = f["mesh"]["array_dimensions"][0] - 2
                else:
                    nrd = f["mesh"]["array_dimensions"][0]

                mask = np.repeat(mask_0[:, :, np.newaxis], nrd, axis=2).transpose()
                specials = (
                    "abar",
                    "e_rms_1",
                    "e_rms_2",
                    "e_rms_3",
                    "e_rms_4",
                    "lumin_1",
                    "lumin_2",
                    "lumin_3",
                    "lumin_4",
                    "num_lumin_1",
                    "num_lumin_2",
                    "num_lumin_3",
                    "num_lumin_4",
                    "shock",
                    "nse_c",
                )
                for field in fields:  # Reads data by locating subheading
                    ftype, fname = field
                    a_name_2 = [i.decode("utf-8") for i in f["abundance"]["a_name"]]
                    a_name_dict = {name.strip(): name for name in a_name_2}
                    if fname not in specials:
                        if fname in f["fluid"]:
                            ds = f["fluid"][f"{fname}"]
                        elif fname in f["abundance"]:
                            ds = f["abundance"][f"{fname}"]
                        elif fname in a_name_dict:
                            ind_xn = a_name_2.index(a_name_dict[fname])
                            ds = f["abundance"]["xn_c"][:, :, :, ind_xn]
                        else:
                            mylog.warning("Invalid field name %s", fname)
                        dat_1 = ds[:, :, :].transpose()

                    elif fname == "nse_c":
                        if np.shape(f["abundance"]["nse_c"]) != np.shape(
                            f["fluid"]["rho_c"]
                        ):
                            ds = f["abundance"]["nse_c"][:, :, 1:]
                        else:
                            ds = f["abundance"]["nse_c"]
                        dat_1 = ds[:, :, :].transpose()
                    elif fname == "abar":
                        xn_c = np.array(f["abundance"]["xn_c"])
                        a_nuc_rep_c = np.array(f["abundance"]["a_nuc_rep_c"])
                        a_nuc = np.array(f["abundance"]["a_nuc"])
                        a_nuc_tile = np.tile(
                            a_nuc, (xn_c.shape[0], xn_c.shape[1], xn_c.shape[2], 1)
                        )
                        yn_c = np.empty(xn_c.shape)
                        yn_c[:, :, :, :-1] = xn_c[:, :, :, :-1] / a_nuc_tile[:, :, :, :]
                        yn_c[:, :, :, -1] = xn_c[:, :, :, -1] / a_nuc_rep_c[:, :, :]
                        ytot = np.sum(yn_c, axis=3)
                        atot = np.sum(xn_c, axis=3)
                        abar = np.divide(atot, ytot)
                        dat_1 = abar[:, :, :].transpose()
                    elif fname in ("e_rms_1", "e_rms_2", "e_rms_3", "e_rms_4"):
                        dims = f["mesh"]["array_dimensions"]
                        n_groups = f["radiation"]["raddim"][0]
                        n_species = f["radiation"]["raddim"][1]
                        n_hyperslabs = f["mesh"]["nz_hyperslabs"][()]
                        energy_edge = f["radiation"]["unubi"][()]
                        energy_center = f["radiation"]["unui"][()]
                        d_energy = []
                        for i in range(0, n_groups):
                            d_energy.append(energy_edge[i + 1] - energy_edge[i])
                        d_energy = np.array(d_energy)
                        e3de = energy_center**3 * d_energy
                        e5de = energy_center**5 * d_energy

                        psi0_c = f["radiation"]["psi0_c"][:]
                        row = np.empty(
                            (n_species, int(dims[2] / n_hyperslabs), dims[1], dims[0])
                        )
                        for n in range(0, n_species):
                            numerator = np.sum(psi0_c[:, :, :, n] * e5de, axis=3)
                            denominator = np.sum(psi0_c[:, :, :, n] * e3de, axis=3)
                            row[n][:][:][:] = np.sqrt(
                                numerator / (denominator + 1e-100)
                            )
                            species = int(fname[-1]) - 1
                            dat_1 = row[species, :, :, :].transpose()
                    elif fname in (
                        "lumin_1",
                        "lumin_2",
                        "lumin_3",
                        "lumin_4",
                        "num_lumin_1",
                        "num_lumin_2",
                        "num_lumin_3",
                        "num_lumin_4",
                    ):
                        dims = f["mesh"]["array_dimensions"]
                        n_groups = f["radiation"]["raddim"][0]
                        n_hyperslabs = f["mesh"]["nz_hyperslabs"][()]
                        ergmev = float((1 * un.MeV) / (1 * un.erg))
                        cvel = float(un.c.to("cm/s"))
                        h = float(un.h.to("MeV * s"))
                        ecoef = 4.0 * np.pi * ergmev / (h * cvel) ** 3
                        radius = f["mesh"]["x_ef"][()]
                        agr_e = f["fluid"]["agr_e"][()]
                        cell_area_GRcorrected = 4 * np.pi * radius**2 / agr_e**4
                        psi1_e = f["radiation"]["psi1_e"]
                        energy_edge = f["radiation"]["unubi"][()]
                        energy_center = f["radiation"]["unui"][()]
                        d_energy = []
                        for i in range(0, n_groups):
                            d_energy.append(energy_edge[i + 1] - energy_edge[i])
                        d_energy = np.array(d_energy)
                        species = int(fname[-1]) - 1
                        if fname in ("lumin_1", "lumin_2", "lumin_3", "lumin_4"):
                            eNde = energy_center**3 * d_energy
                        else:
                            eNde = energy_center**2 * d_energy

                        lumin = (
                            np.sum(psi1_e[:, :, :, species] * eNde, axis=3)
                            * np.tile(
                                cell_area_GRcorrected[1 : dims[0] + 1],
                                (int(dims[2] / n_hyperslabs), dims[1], 1),
                            )
                            * (cvel * ecoef * 1e-51)
                        )
                        dat_1 = lumin[:, :, :].transpose()

                    if f["mesh"]["array_dimensions"][2] > 1:
                        data = dat_1[:-2, :, :]  # Clips off ghost zones for 3D
                    else:
                        data = dat_1[:, :, :]
                        data = np.ma.masked_where(mask == 0.0, data)  # Masks
                        data = np.ma.filled(
                            data, fill_value=0.0
                        )  # Replaces masked value with 0

                    yield field, chunk.objs[0], data

        pass

    def _read_chunk_data(self, chunk, fields):
        # This reads the data from a single chunk without doing any selection,
        # and is only used for caching data that might be used by multiple
        # different selectors later. For instance, this can speed up ghost zone
        # computation.
        pass



# distutils: include_dirs = EMBREE_INC_DIR
# distutils: library_dirs = EMBREE_LIB_DIR
# distutils: libraries = EMBREE_LIBS
# distutils: language = c++
"""
This file contains functions used for performing ray-tracing with Embree
for 2nd-order Lagrange Elements.

Note - this file is only used for the Embree-accelerated ray-tracer.

"""


cimport cython
cimport numpy as np
cimport pyembree.rtcore as rtc
cimport pyembree.rtcore_geometry as rtcg
cimport pyembree.rtcore_ray as rtcr
from libc.math cimport fabs, fmax, fmin, sqrt
from pyembree.rtcore cimport Vec3f

from yt.utilities.lib.bounding_volume_hierarchy cimport BBox
from yt.utilities.lib.primitives cimport (
    RayHitData,
    compute_patch_hit,
    compute_tet_patch_hit,
    patchSurfaceDerivU,
    patchSurfaceDerivV,
    patchSurfaceFunc,
    tet_patchSurfaceDerivU,
    tet_patchSurfaceDerivV,
    tet_patchSurfaceFunc,
)
from yt.utilities.lib.vec3_ops cimport cross, distance, dot, subtract

from .mesh_samplers cimport sample_hex20, sample_tet10


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void patchBoundsFunc(Patch* patches,
                          size_t item,
                          rtcg.RTCBounds* bounds_o) nogil:

    cdef Patch patch = patches[item]

    cdef float lo_x = 1.0e300
    cdef float lo_y = 1.0e300
    cdef float lo_z = 1.0e300

    cdef float hi_x = -1.0e300
    cdef float hi_y = -1.0e300
    cdef float hi_z = -1.0e300

    cdef int i
    for i in range(8):
        lo_x = fmin(lo_x, patch.v[i][0])
        lo_y = fmin(lo_y, patch.v[i][1])
        lo_z = fmin(lo_z, patch.v[i][2])
        hi_x = fmax(hi_x, patch.v[i][0])
        hi_y = fmax(hi_y, patch.v[i][1])
        hi_z = fmax(hi_z, patch.v[i][2])

    bounds_o.lower_x = lo_x
    bounds_o.lower_y = lo_y
    bounds_o.lower_z = lo_z
    bounds_o.upper_x = hi_x
    bounds_o.upper_y = hi_y
    bounds_o.upper_z = hi_z


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void patchIntersectFunc(Patch* patches,
                             rtcr.RTCRay& ray,
                             size_t item) nogil:

    cdef Patch patch = patches[item]

    cdef RayHitData hd = compute_patch_hit(patch.v, ray.org, ray.dir)

    # only count this is it's the closest hit
    if (hd.t < ray.tnear or hd.t > ray.Ng[0]):
        return

    if (fabs(hd.u) <= 1.0 and fabs(hd.v) <= 1.0 and hd.converged):

        # we have a hit, so update ray information
        ray.u = hd.u
        ray.v = hd.v
        ray.geomID = patch.geomID
        ray.primID = item
        ray.Ng[0] = hd.t

        # sample the solution at the calculated point
        sample_hex20(patches, ray)

    return

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void tet_patchBoundsFunc(Tet_Patch* tet_patches,
                          size_t item,
                          rtcg.RTCBounds* bounds_o) nogil:

    cdef Tet_Patch tet_patch = tet_patches[item]

    cdef float lo_x = 1.0e300
    cdef float lo_y = 1.0e300
    cdef float lo_z = 1.0e300

    cdef float hi_x = -1.0e300
    cdef float hi_y = -1.0e300
    cdef float hi_z = -1.0e300

    cdef int i
    for i in range(6):
        lo_x = fmin(lo_x, tet_patch.v[i][0])
        lo_y = fmin(lo_y, tet_patch.v[i][1])
        lo_z = fmin(lo_z, tet_patch.v[i][2])
        hi_x = fmax(hi_x, tet_patch.v[i][0])
        hi_y = fmax(hi_y, tet_patch.v[i][1])
        hi_z = fmax(hi_z, tet_patch.v[i][2])

    bounds_o.lower_x = lo_x
    bounds_o.lower_y = lo_y
    bounds_o.lower_z = lo_z
    bounds_o.upper_x = hi_x
    bounds_o.upper_y = hi_y
    bounds_o.upper_z = hi_z


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void tet_patchIntersectFunc(Tet_Patch* tet_patches,
                             rtcr.RTCRay& ray,
                             size_t item) nogil:

    cdef Tet_Patch tet_patch = tet_patches[item]

    cdef RayHitData hd = compute_tet_patch_hit(tet_patch.v, ray.org, ray.dir)

    # only count this is it's the closest hit
    if (hd.t < ray.tnear or hd.t > ray.Ng[0]):
        return

    if (hd.converged and 0 <= hd.u and 0 <= hd.v and hd.u + hd.v <= 1):

        # we have a hit, so update ray information
        ray.u = hd.u
        ray.v = hd.v
        ray.geomID = tet_patch.geomID
        ray.primID = item
        ray.Ng[0] = hd.t

        # sample the solution at the calculated point
        sample_tet10(tet_patches, ray)

    return



"""
Definitions for the partitioned grid




"""


import numpy as np

cimport cython
cimport numpy as np

from .volume_container cimport VolumeContainer


cdef class PartitionedGrid:
    cdef public object my_data
    cdef public object source_mask
    cdef public object LeftEdge
    cdef public object RightEdge
    cdef public int parent_grid_id
    cdef VolumeContainer *container
    cdef np.float64_t star_er
    cdef np.float64_t star_sigma_num
    cdef np.float64_t star_coeff
    cdef void get_vector_field(self, np.float64_t pos[3],
                               np.float64_t *vel, np.float64_t *vel_mag)



# distutils: libraries = STD_LIBS
cimport numpy as np

import cython
import numpy as np

from libc.stdio cimport *

import struct


cdef INT32_SIZE = sizeof(np.int32_t)
cdef DOUBLE_SIZE = sizeof(np.float64_t)

cdef class FortranFile:
    """This class provides facilities to interact with files written
    in fortran-record format.  Since this is a non-standard file
    format, whose contents depend on the compiler and the endianness
    of the machine, caution is advised. This code will assume that the
    record header is written as a 32bit (4byte) signed integer. The
    code also assumes that the records use the system's local
    endianness.

    Notes
    -----
    Since the assumed record header is an signed integer on 32bit, it
    will overflow at 2**31=2147483648 elements.

    This module has been inspired by scipy's FortranFile, especially
    the docstrings.
    """
    def __cinit__(self, str fname):
        self.cfile = fopen(fname.encode('utf-8'), 'rb')
        self._closed = False

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    cpdef INT64_t skip(self, INT64_t n=1) except -1:
        """Skip records.

        Parameters
        ----------
        - n : integer
            The number of records to skip

        Returns
        -------
        value : int
            Returns 0 on success.
        """
        cdef INT32_t s1, s2, i

        if self._closed:
            raise ValueError("Read of closed file.")

        for i in range(n):
            fread(&s1, INT32_SIZE, 1, self.cfile)
            fseek(self.cfile, s1, SEEK_CUR)
            fread(&s2, INT32_SIZE, 1, self.cfile)

            if s1 != s2:
                raise IOError('Sizes do not agree in the header and footer for '
                              'this record - check header dtype. Got %s and %s' % (s1, s2))

        return 0

    cdef INT64_t get_size(self, str dtype):
        """Return the size of an element given its datatype.

        Parameters
        ----------
        dtype : str
           The dtype, see note for details about the values of dtype.

        Returns
        -------
        size : int
           The size in byte of the dtype

        Note:
        -----
        See
        https://docs.python.org/3.5/library/struct.html#format-characters
        for details about the formatting characters.
        """
        if dtype == 'i':
            return 4
        elif dtype == 'd':
            return 8
        elif dtype == 'f':
            return 4
        elif dtype == 'l':
            return 8
        else:
            # Fallback to (slow) numpy-based to compute the size
            return np.dtype(dtype).itemsize

    cpdef np.ndarray read_vector(self, str dtype):
        """Reads a record from the file and return it as numpy array.

        Parameters
        ----------
        d : data type
            This is the datatype (from the struct module) that we should read.

        Returns
        -------
        tr : numpy.ndarray
            This is the vector of values read from the file.

        Examples
        --------
        >>> f = FortranFile("fort.3")
        >>> rv = f.read_vector("d")  # Read a float64 array
        >>> rv = f.read_vector("i")  # Read an int32 array
        """
        cdef INT32_t s1, s2, size
        cdef np.ndarray data

        if self._closed:
            raise ValueError("I/O operation on closed file.")

        size = self.get_size(dtype)

        fread(&s1, INT32_SIZE, 1, self.cfile)

        # Check record is compatible with data type
        if s1 % size != 0:
            raise ValueError('Size obtained (%s) does not match with the expected '
                             'size (%s) of multi-item record' % (s1, size))

        data = np.empty(s1 // size, dtype=dtype)
        fread(<void *>data.data, size, s1 // size, self.cfile)
        fread(&s2, INT32_SIZE, 1, self.cfile)

        if s1 != s2:
            raise IOError('Sizes do not agree in the header and footer for '
                          'this record - check header dtype')

        return data

    cpdef INT32_t read_int(self) except? -1:
        """Reads a single int32 from the file and return it.

        Returns
        -------
        data : int32
            The value.

        Examples
        --------
        >>> f = FortranFile("fort.3")
        >>> rv = f.read_vector("d")  # Read a float64 array
        >>> rv = f.read_vector("i")  # Read an int32 array
        """

        cdef INT32_t s1, s2
        cdef INT32_t data

        if self._closed:
            raise ValueError("I/O operation on closed file.")

        fread(&s1, INT32_SIZE, 1, self.cfile)

        if s1 != INT32_SIZE != 0:
            raise ValueError('Size obtained (%s) does not match with the expected '
                             'size (%s) of record' % (s1, INT32_SIZE))

        fread(&data, INT32_SIZE, s1 // INT32_SIZE, self.cfile)
        fread(&s2, INT32_SIZE, 1, self.cfile)

        if s1 != s2:
            raise IOError('Sizes do not agree in the header and footer for '
                          'this record - check header dtype')

        return data

    def read_attrs(self, object attrs):
        """This function reads from that file according to a
        definition of attributes, returning a dictionary.

        Fortran unformatted files provide total bytesize at the
        beginning and end of a record. By correlating the components
        of that record with attribute names, we construct a dictionary
        that gets returned. Note that this function is used for
        reading sequentially-written records. If you have many written
        that were written simultaneously.

        Parameters
        ----------
        attrs : iterable of iterables
            This object should be an iterable of one of the formats:
            [ (attr_name, count, struct type), ... ].
            [ ((name1,name2,name3), count, vector type]
            [ ((name1,name2,name3), count, 'type type type']
            [ (attr_name, count, struct type, optional)]

            `optional` : boolean.
                If True, the attribute can be stored as an empty Fortran record.

        Returns
        -------
        values : dict
            This will return a dict of iterables of the components of the values in
            the file.

        Examples
        --------

        >>> header = [ ("ncpu", 1, "i"), ("nfiles", 2, "i") ]
        >>> f = FortranFile("fort.3")
        >>> rv = f.read_attrs(header)
        """

        cdef str dtype
        cdef int n
        cdef dict data
        cdef key
        cdef np.ndarray tmp
        cdef bint optional

        if self._closed:
            raise ValueError("I/O operation on closed file.")

        data = {}

        for a in attrs:
            if len(a) == 3:
                key, n, dtype = a
                optional = False
            else:
                key, n, dtype, optional = a
            if n == 1:
                tmp = self.read_vector(dtype)
                if len(tmp) == 0 and optional:
                    continue
                elif (len(tmp) == 1) or (n == -1):
                    data[key] = tmp[0]
                else:
                    raise ValueError("Expected a record of length %s, got %s (%s)" % (n, len(tmp), key))
            else:
                tmp = self.read_vector(dtype)
                if (len(tmp) == 0 and optional):
                    continue
                elif (len(tmp) != n) and (n != -1):
                    raise ValueError("Expected a record of length %s, got %s (%s)" % (n, len(tmp), key))

                if isinstance(key, tuple):
                    # There are multiple keys
                    for ikey in range(n):
                        data[key[ikey]] = tmp[ikey]
                else:
                    data[key] = tmp

        return data

    cpdef INT64_t tell(self) except -1:
        """Return current stream position."""
        cdef INT64_t pos

        if self._closed:
            raise ValueError("I/O operation on closed file.")

        pos = ftell(self.cfile)
        return pos

    cpdef INT64_t seek(self, INT64_t pos, INT64_t whence=SEEK_SET) except -1:
        """Change stream position.

        Parameters
        ----------
        pos : int
            Change the stream position to the given byte offset. The offset is
            interpreted relative to the position indicated by whence.
        whence : int
            Determine how pos is interpreted. Can by any of
            * 0 -- start of stream (the default); offset should be zero or positive
            * 1 -- current stream position; offset may be negative
            * 2 -- end of stream; offset is usually negative

        Returns
        -------
        pos : int
            The new absolute position.
        """
        if self._closed:
            raise ValueError("I/O operation on closed file.")
        if whence < 0 or whence > 2:
            raise ValueError("whence argument can be 0, 1, or 2. Got %s" % whence)

        fseek(self.cfile, pos, whence)
        return self.tell()

    cpdef void close(self):
        """Close the file descriptor.

        This method has no effect if the file is already closed.
        """
        if self._closed:
            return
        fclose(self.cfile)
        self._closed = True

    def __dealloc__(self):
        if not self._closed:
            self.close()



import numpy as np

from yt.utilities.file_handler import NetCDF4FileHandler
from yt.utilities.io_handler import BaseIOHandler


class CM1IOHandler(BaseIOHandler):
    _particle_reader = False
    _dataset_type = "cm1"

    def __init__(self, ds):
        self.filename = ds.filename
        self._handle = NetCDF4FileHandler(self.filename)
        super().__init__(ds)

    def _read_particle_coords(self, chunks, ptf):
        # This needs to *yield* a series of tuples of (ptype, (x, y, z)).
        # chunks is a list of chunks, and ptf is a dict where the keys are
        # ptypes and the values are lists of fields.
        raise NotImplementedError

    def _read_particle_fields(self, chunks, ptf, selector):
        # This gets called after the arrays have been allocated.  It needs to
        # yield ((ptype, field), data) where data is the masked results of
        # reading ptype, field and applying the selector to the data read in.
        # Selector objects have a .select_points(x,y,z) that returns a mask, so
        # you need to do your masking here.
        raise NotImplementedError

    def _read_fluid_selection(self, chunks, selector, fields, size):
        # This needs to allocate a set of arrays inside a dictionary, where the
        # keys are the (ftype, fname) tuples and the values are arrays that
        # have been masked using whatever selector method is appropriate.  The
        # dict gets returned at the end and it should be flat, with selected
        # data.  Note that if you're reading grid data, you might need to
        # special-case a grid selector object.
        # Also note that "chunks" is a generator for multiple chunks, each of
        # which contains a list of grids. The returned numpy arrays should be
        # in 64-bit float and contiguous along the z direction. Therefore, for
        # a C-like input array with the dimension [x][y][z] or a
        # Fortran-like input array with the dimension (z,y,x), a matrix
        # transpose is required (e.g., using np_array.transpose() or
        # np_array.swapaxes(0,2)).

        data = {}
        offset = 0
        with self._handle.open_ds() as ds:
            for field in fields:
                data[field] = np.empty(size, dtype="float64")
                for chunk in chunks:
                    for grid in chunk.objs:
                        variable = ds.variables[field[1]][:][0]
                        values = np.squeeze(variable.T)
                        offset += grid.select(selector, values, data[field], offset)
        return data

    def _read_chunk_data(self, chunk, fields):
        # This reads the data from a single chunk without doing any selection,
        # and is only used for caching data that might be used by multiple
        # different selectors later. For instance, this can speed up ghost zone
        # computation.
        pass



from yt.frontends.sph.fields import SPHFieldInfo


class TipsyFieldInfo(SPHFieldInfo):
    known_particle_fields = SPHFieldInfo.known_particle_fields + (
        ("smoothing_length", ("code_length", [], None)),
    )
    aux_particle_fields = {
        "uDotFB": ("uDotFB", ("code_mass * code_velocity**2", [""], None)),
        "uDotAV": ("uDotAV", ("code_mass * code_velocity**2", [""], None)),
        "uDotPdV": ("uDotPdV", ("code_mass * code_velocity**2", [""], None)),
        "uDotHydro": ("uDotHydro", ("code_mass * code_velocity**2", [""], None)),
        "uDotDiff": ("uDotDiff", ("code_mass * code_velocity**2", [""], None)),
        "uDot": ("uDot", ("code_mass * code_velocity**2", [""], None)),
        "coolontime": ("coolontime", ("code_time", [""], None)),
        "timeform": ("timeform", ("code_time", [""], None)),
        "massform": ("massform", ("code_mass", [""], None)),
        "HI": ("HI", ("dimensionless", ["H_fraction"], None)),
        "HII": ("HII", ("dimensionless", ["H_p1_fraction"], None)),
        "HeI": ("HeI", ("dimensionless", ["He_fraction"], None)),
        "HeII": ("HeII", ("dimensionless", ["He_p2_fraction"], None)),
        "OxMassFrac": ("OxMassFrac", ("dimensionless", ["O_fraction"], None)),
        "FeMassFrac": ("FeMassFrac", ("dimensionless", ["Fe_fraction"], None)),
        "c": ("c", ("code_velocity", [""], None)),
        "acc": ("acc", ("code_velocity / code_time", [""], None)),
        "accg": ("accg", ("code_velocity / code_time", [""], None)),
        "smoothlength": ("smoothlength", ("code_length", ["smoothing_length"], None)),
    }

    def __init__(self, ds, field_list, slice_info=None):
        for field in field_list:
            if (
                field[1] in self.aux_particle_fields.keys()
                and self.aux_particle_fields[field[1]] not in self.known_particle_fields
            ):
                self.known_particle_fields += (self.aux_particle_fields[field[1]],)
        super().__init__(ds, field_list, slice_info)



from . import tests
from .data_structures import (
    GadgetFOFDataset,
    GadgetFOFHaloContainer,
    GadgetFOFHaloDataset,
    GadgetFOFHaloParticleIndex,
    GadgetFOFHDF5File,
    GadgetFOFParticleIndex,
)
from .fields import GadgetFOFFieldInfo, GadgetFOFHaloFieldInfo
from .io import IOHandlerGadgetFOFHaloHDF5, IOHandlerGadgetFOFHDF5



import numpy as np

from yt.geometry.fake_octree import create_fake_octree
from yt.geometry.oct_container import ParticleOctreeContainer, RAMSESOctreeContainer

nocts = 3
max_level = 12
dn = 2
dd = np.ones(3, dtype="i4") * dn
dle = np.ones(3, dtype="f8") * 0.0
dre = np.ones(3, dtype="f8")
fsub = 0.25
domain = 1

oct_handler = RAMSESOctreeContainer(dd, dle, dre)
leaves = create_fake_octree(oct_handler, nocts, max_level, dd, dle, dre, fsub)
mask = np.ones((nocts, 8), dtype="bool")
cell_count = nocts * 8
oct_counts = oct_handler.count_levels(max_level, 1, mask)
level_counts = np.concatenate(
    (
        [
            0,
        ],
        np.cumsum(oct_counts),
    )
)
fc = oct_handler.fcoords(domain, mask, cell_count, level_counts.copy())
leavesb = oct_handler.count_leaves(mask)
assert leaves == leavesb

# Now take the fcoords, call them particles and recreate the same octree
print("particle-based recreate")
oct_handler2 = ParticleOctreeContainer(dd, dle, dre)
oct_handler2.allocate_domains([nocts])
oct_handler2.n_ref = 1  # specifically make a maximum of 1 particle per oct
oct_handler2.add(fc, 1)
print("added particles")
cell_count2 = nocts * 8
oct_counts2 = oct_handler.count_levels(max_level, 1, mask)
level_counts2 = np.concatenate(
    (
        [
            0,
        ],
        np.cumsum(oct_counts),
    )
)
fc2 = oct_handler.fcoords(domain, mask, cell_count, level_counts.copy())
leaves2 = oct_handler2.count_leaves(mask)
assert leaves == leaves2

print("success")



# This module is not part of the public namespace `yt.units`
# It is home to wrapper functions that are directly copied from unyt 2.9.2
# We vendor them as a transition step towards unyt 3.0 (in devlopment),
# where these wrapper functions are deprecated and are should be replaced with vanilla numpy API
# FUTURE:
# - require unyt>=3.0
# - deprecate these functions in yt too

from unyt import unyt_array, unyt_quantity
import numpy as np


def _validate_numpy_wrapper_units(v, arrs):
    if not any(isinstance(a, unyt_array) for a in arrs):
        return v
    if not all(isinstance(a, unyt_array) for a in arrs):
        raise RuntimeError("Not all of your arrays are unyt_arrays.")
    a1 = arrs[0]
    if not all(a.units == a1.units for a in arrs[1:]):
        raise RuntimeError("Your arrays must have identical units.")
    v.units = a1.units
    return v


def uconcatenate(arrs, axis=0):
    """Concatenate a sequence of arrays.

    This wrapper around numpy.concatenate preserves units. All input arrays
    must have the same units.  See the documentation of numpy.concatenate for
    full details.

    Examples
    --------
    >>> from unyt import cm
    >>> A = [1, 2, 3]*cm
    >>> B = [2, 3, 4]*cm
    >>> uconcatenate((A, B))
    unyt_array([1, 2, 3, 2, 3, 4], 'cm')

    """
    v = np.concatenate(arrs, axis=axis)
    v = _validate_numpy_wrapper_units(v, arrs)
    return v


def ucross(arr1, arr2, registry=None, axisa=-1, axisb=-1, axisc=-1, axis=None):
    """Applies the cross product to two YT arrays.

    This wrapper around numpy.cross preserves units.
    See the documentation of numpy.cross for full
    details.
    """
    v = np.cross(arr1, arr2, axisa=axisa, axisb=axisb, axisc=axisc, axis=axis)
    units = arr1.units * arr2.units
    arr = unyt_array(v, units, registry=registry)
    return arr


def uintersect1d(arr1, arr2, assume_unique=False):
    """Find the sorted unique elements of the two input arrays.

    A wrapper around numpy.intersect1d that preserves units.  All input arrays
    must have the same units.  See the documentation of numpy.intersect1d for
    full details.

    Examples
    --------
    >>> from unyt import cm
    >>> A = [1, 2, 3]*cm
    >>> B = [2, 3, 4]*cm
    >>> uintersect1d(A, B)
    unyt_array([2, 3], 'cm')

    """
    v = np.intersect1d(arr1, arr2, assume_unique=assume_unique)
    v = _validate_numpy_wrapper_units(v, [arr1, arr2])
    return v


def uunion1d(arr1, arr2):
    """Find the union of two arrays.

    A wrapper around numpy.intersect1d that preserves units.  All input arrays
    must have the same units.  See the documentation of numpy.intersect1d for
    full details.

    Examples
    --------
    >>> from unyt import cm
    >>> A = [1, 2, 3]*cm
    >>> B = [2, 3, 4]*cm
    >>> uunion1d(A, B)
    unyt_array([1, 2, 3, 4], 'cm')

    """
    v = np.union1d(arr1, arr2)
    v = _validate_numpy_wrapper_units(v, [arr1, arr2])
    return v


def unorm(data, ord=None, axis=None, keepdims=False):
    """Matrix or vector norm that preserves units

    This is a wrapper around np.linalg.norm that preserves units. See
    the documentation for that function for descriptions of the keyword
    arguments.

    Examples
    --------
    >>> from unyt import km
    >>> data = [1, 2, 3]*km
    >>> print(unorm(data))
    3.7416573867739413 km
    """
    norm = np.linalg.norm(data, ord=ord, axis=axis, keepdims=keepdims)
    if norm.shape == ():
        return unyt_quantity(norm, data.units)
    return unyt_array(norm, data.units)


def udot(op1, op2):
    """Matrix or vector dot product that preserves units

    This is a wrapper around np.dot that preserves units.

    Examples
    --------
    >>> from unyt import km, s
    >>> a = np.eye(2)*km
    >>> b = (np.ones((2, 2)) * 2)*s
    >>> print(udot(a, b))
    [[2. 2.]
     [2. 2.]] km*s
    """
    dot = np.dot(op1.d, op2.d)
    units = op1.units * op2.units
    if dot.shape == ():
        return unyt_quantity(dot, units)
    return unyt_array(dot, units)


def uvstack(arrs):
    """Stack arrays in sequence vertically (row wise) while preserving units

    This is a wrapper around np.vstack that preserves units.

    Examples
    --------
    >>> from unyt import km
    >>> a = [1, 2, 3]*km
    >>> b = [2, 3, 4]*km
    >>> print(uvstack([a, b]))
    [[1 2 3]
     [2 3 4]] km
    """
    v = np.vstack(arrs)
    v = _validate_numpy_wrapper_units(v, arrs)
    return v


def uhstack(arrs):
    """Stack arrays in sequence horizontally while preserving units

    This is a wrapper around np.hstack that preserves units.

    Examples
    --------
    >>> from unyt import km
    >>> a = [1, 2, 3]*km
    >>> b = [2, 3, 4]*km
    >>> print(uhstack([a, b]))
    [1 2 3 2 3 4] km
    >>> a = [[1],[2],[3]]*km
    >>> b = [[2],[3],[4]]*km
    >>> print(uhstack([a, b]))
    [[1 2]
     [2 3]
     [3 4]] km
    """
    v = np.hstack(arrs)
    v = _validate_numpy_wrapper_units(v, arrs)
    return v


def ustack(arrs, axis=0):
    """Join a sequence of arrays along a new axis while preserving units

    The axis parameter specifies the index of the new axis in the
    dimensions of the result. For example, if ``axis=0`` it will be the
    first dimension and if ``axis=-1`` it will be the last dimension.

    This is a wrapper around np.stack that preserves units. See the
    documentation for np.stack for full details.

    Examples
    --------
    >>> from unyt import km
    >>> a = [1, 2, 3]*km
    >>> b = [2, 3, 4]*km
    >>> print(ustack([a, b]))
    [[1 2 3]
     [2 3 4]] km
    """
    v = np.stack(arrs, axis=axis)
    v = _validate_numpy_wrapper_units(v, arrs)
    return v



import numpy as np

from yt.testing import assert_array_almost_equal
from yt.utilities.math_utils import (
    get_cyl_r,
    get_cyl_r_component,
    get_cyl_theta,
    get_cyl_theta_component,
    get_cyl_z,
    get_cyl_z_component,
    get_sph_phi,
    get_sph_phi_component,
    get_sph_r,
    get_sph_r_component,
    get_sph_theta,
    get_sph_theta_component,
)

# Randomly generated coordinates in the domain [[-1,1],[-1,1],-1,1]]
coords = np.array(
    [
        [-0.41503037, -0.22102472, -0.55774212],
        [0.73828247, -0.17913899, 0.64076921],
        [0.08922066, -0.94254844, -0.61774511],
        [0.10173242, -0.95789145, 0.16294352],
        [0.73186508, -0.3109153, 0.75728738],
        [0.8757989, -0.41475119, -0.57039201],
        [0.58040762, 0.81969082, 0.46759728],
        [-0.89983356, -0.9853683, -0.38355343],
    ]
).T


def test_spherical_coordinate_conversion():
    normal = [0, 0, 1]
    real_r = [
        0.72950559,
        0.99384957,
        1.13047198,
        0.97696269,
        1.09807968,
        1.12445067,
        1.10788685,
        1.38843954,
    ]
    real_theta = [
        2.44113629,
        0.87012028,
        2.14891444,
        1.4032274,
        0.80979483,
        2.10280198,
        1.13507735,
        1.85068416,
    ]
    real_phi = [
        -2.65224483,
        -0.23804243,
        -1.47641858,
        -1.46498842,
        -0.40172325,
        -0.4422801,
        0.95466734,
        -2.31085392,
    ]

    calc_r = get_sph_r(coords)
    calc_theta = get_sph_theta(coords, normal)
    calc_phi = get_sph_phi(coords, normal)

    assert_array_almost_equal(calc_r, real_r)
    assert_array_almost_equal(calc_theta, real_theta)
    assert_array_almost_equal(calc_phi, real_phi)

    normal = [1, 0, 0]
    real_theta = [
        2.17598842,
        0.73347681,
        1.49179079,
        1.46647589,
        0.8412984,
        0.67793705,
        1.0193883,
        2.27586987,
    ]
    real_phi = [
        -1.94809584,
        1.843405,
        -2.56143151,
        2.97309903,
        1.96037671,
        -2.1995016,
        0.51841239,
        -2.77038877,
    ]

    calc_theta = get_sph_theta(coords, normal)
    calc_phi = get_sph_phi(coords, normal)

    assert_array_almost_equal(calc_theta, real_theta)
    assert_array_almost_equal(calc_phi, real_phi)


def test_cylindrical_coordinate_conversion():
    normal = [0, 0, 1]
    real_r = [
        0.47021498,
        0.75970506,
        0.94676179,
        0.96327853,
        0.79516968,
        0.96904193,
        1.00437346,
        1.3344104,
    ]
    real_theta = [
        -2.65224483,
        -0.23804243,
        -1.47641858,
        -1.46498842,
        -0.40172325,
        -0.4422801,
        0.95466734,
        -2.31085392,
    ]
    real_z = [
        -0.55774212,
        0.64076921,
        -0.61774511,
        0.16294352,
        0.75728738,
        -0.57039201,
        0.46759728,
        -0.38355343,
    ]

    calc_r = get_cyl_r(coords, normal)
    calc_theta = get_cyl_theta(coords, normal)
    calc_z = get_cyl_z(coords, normal)

    assert_array_almost_equal(calc_r, real_r)
    assert_array_almost_equal(calc_theta, real_theta)
    assert_array_almost_equal(calc_z, real_z)

    normal = [1, 0, 0]
    real_r = [
        0.59994016,
        0.66533898,
        1.12694569,
        0.97165149,
        0.81862843,
        0.70524152,
        0.94368441,
        1.05738542,
    ]
    real_theta = [
        -1.94809584,
        1.843405,
        -2.56143151,
        2.97309903,
        1.96037671,
        -2.1995016,
        0.51841239,
        -2.77038877,
    ]
    real_z = [
        -0.41503037,
        0.73828247,
        0.08922066,
        0.10173242,
        0.73186508,
        0.8757989,
        0.58040762,
        -0.89983356,
    ]

    calc_r = get_cyl_r(coords, normal)
    calc_theta = get_cyl_theta(coords, normal)
    calc_z = get_cyl_z(coords, normal)

    assert_array_almost_equal(calc_r, real_r)
    assert_array_almost_equal(calc_theta, real_theta)
    assert_array_almost_equal(calc_z, real_z)


def test_spherical_coordinate_projections():
    normal = [0, 0, 1]
    theta = get_sph_theta(coords, normal)
    phi = get_sph_phi(coords, normal)
    zero = np.tile(0, coords.shape[1])

    # Purely radial field
    vecs = np.array(
        [np.sin(theta) * np.cos(phi), np.sin(theta) * np.sin(phi), np.cos(theta)]
    )
    assert_array_almost_equal(zero, get_sph_theta_component(vecs, theta, phi, normal))
    assert_array_almost_equal(zero, get_sph_phi_component(vecs, phi, normal))

    # Purely toroidal field
    vecs = np.array([-np.sin(phi), np.cos(phi), zero])
    assert_array_almost_equal(zero, get_sph_theta_component(vecs, theta, phi, normal))
    assert_array_almost_equal(zero, get_sph_r_component(vecs, theta, phi, normal))

    # Purely poloidal field
    vecs = np.array(
        [np.cos(theta) * np.cos(phi), np.cos(theta) * np.sin(phi), -np.sin(theta)]
    )
    assert_array_almost_equal(zero, get_sph_phi_component(vecs, phi, normal))
    assert_array_almost_equal(zero, get_sph_r_component(vecs, theta, phi, normal))


def test_cylindrical_coordinate_projections():
    normal = [0, 0, 1]
    theta = get_cyl_theta(coords, normal)
    z = get_cyl_z(coords, normal)
    zero = np.tile(0, coords.shape[1])

    # Purely radial field
    vecs = np.array([np.cos(theta), np.sin(theta), zero])
    assert_array_almost_equal(zero, get_cyl_theta_component(vecs, theta, normal))
    assert_array_almost_equal(zero, get_cyl_z_component(vecs, normal))

    # Purely toroidal field
    vecs = np.array([-np.sin(theta), np.cos(theta), zero])
    assert_array_almost_equal(zero, get_cyl_z_component(vecs, normal))
    assert_array_almost_equal(zero, get_cyl_r_component(vecs, theta, normal))

    # Purely z field
    vecs = np.array([zero, zero, z])
    assert_array_almost_equal(zero, get_cyl_theta_component(vecs, theta, normal))
    assert_array_almost_equal(zero, get_cyl_r_component(vecs, theta, normal))



import numpy as np

from yt.funcs import mylog
from yt.geometry.selection_routines import GridSelector
from yt.utilities.io_handler import BaseParticleIOHandler
from yt.utilities.on_demand_imports import _h5py as h5py


def _grid_dname(grid_id):
    return "/data/grid_%010i" % grid_id


def _field_dname(grid_id, field_name):
    return f"{_grid_dname(grid_id)}/{field_name}"


# TODO all particle bits were removed
class IOHandlerGDFHDF5(BaseParticleIOHandler):
    _dataset_type = "grid_data_format"
    _offset_string = "data:offsets=0"
    _data_string = "data:datatype=0"

    def _read_fluid_selection(self, chunks, selector, fields, size):

        rv = {}
        chunks = list(chunks)

        if isinstance(selector, GridSelector):
            if not (len(chunks) == len(chunks[0].objs) == 1):
                raise RuntimeError
            grid = chunks[0].objs[0]
            h5f = h5py.File(grid.filename, mode="r")
            gds = h5f.get(_grid_dname(grid.id))
            for ftype, fname in fields:
                if self.ds.field_ordering == 1:
                    rv[(ftype, fname)] = gds.get(fname)[()].swapaxes(0, 2)
                else:
                    rv[(ftype, fname)] = gds.get(fname)[()]
            h5f.close()
            return rv
        if size is None:
            size = sum(grid.count(selector) for chunk in chunks for grid in chunk.objs)

        if any((ftype != "gdf" for ftype, fname in fields)):
            raise NotImplementedError

        for field in fields:
            ftype, fname = field
            fsize = size
            # check the dtype instead
            rv[field] = np.empty(fsize, dtype="float64")
        ngrids = sum(len(chunk.objs) for chunk in chunks)
        mylog.debug(
            "Reading %s cells of %s fields in %s blocks",
            size,
            [fn for ft, fn in fields],
            ngrids,
        )
        ind = 0
        for chunk in chunks:
            fid = None
            for grid in chunk.objs:
                if grid.filename is None:
                    continue
                if fid is None:
                    fid = h5py.h5f.open(
                        bytes(grid.filename, "utf-8"), h5py.h5f.ACC_RDONLY
                    )
                if self.ds.field_ordering == 1:
                    # check the dtype instead
                    data = np.empty(grid.ActiveDimensions[::-1], dtype="float64")
                    data_view = data.swapaxes(0, 2)
                else:
                    # check the dtype instead
                    data_view = data = np.empty(grid.ActiveDimensions, dtype="float64")
                for field in fields:
                    ftype, fname = field
                    dg = h5py.h5d.open(
                        fid, bytes(_field_dname(grid.id, fname), "utf-8")
                    )
                    dg.read(h5py.h5s.ALL, h5py.h5s.ALL, data)
                    # caches
                    nd = grid.select(selector, data_view, rv[field], ind)
                ind += nd  # I don't get that part, only last nd is added
            if fid is not None:
                fid.close()
        return rv



import os
import re
import weakref

import numpy as np

from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.static_output import Dataset
from yt.fields.magnetic_field import get_magnetic_normalization
from yt.funcs import mylog, sglob
from yt.geometry.geometry_handler import YTDataChunk
from yt.geometry.grid_geometry_handler import GridIndex
from yt.utilities.chemical_formulas import compute_mu
from yt.utilities.decompose import decompose_array, get_psize
from yt.utilities.lib.misc_utilities import get_box_grids_level

from .fields import AthenaFieldInfo


def chk23(strin):
    return strin.encode("utf-8")


def str23(strin):
    if isinstance(strin, list):
        return [s.decode("utf-8") for s in strin]
    else:
        return strin.decode("utf-8")


def check_readline(fl):
    line = fl.readline()
    chk = chk23("SCALARS")
    if chk in line and not line.startswith(chk):
        line = line[line.find(chk) :]
    chk = chk23("VECTORS")
    if chk in line and not line.startswith(chk):
        line = line[line.find(chk) :]
    return line


def check_break(line):
    splitup = line.strip().split()
    do_break = chk23("SCALAR") in splitup
    do_break = (chk23("VECTOR") in splitup) & do_break
    do_break = (chk23("TABLE") in splitup) & do_break
    do_break = (len(line) == 0) & do_break
    return do_break


def _get_convert(fname):
    def _conv(data):
        return data.convert(fname)

    return _conv


class AthenaGrid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, id, index, level, start, dimensions, file_offset, read_dims):
        gname = index.grid_filenames[id]
        AMRGridPatch.__init__(self, id, filename=gname, index=index)
        self.filename = gname
        self.Parent = []
        self.Children = []
        self.Level = level
        self.start_index = start.copy()
        self.stop_index = self.start_index + dimensions
        self.ActiveDimensions = dimensions.copy()
        self.file_offset = file_offset
        self.read_dims = read_dims

    def _setup_dx(self):
        # So first we figure out what the index is.  We don't assume
        # that dx=dy=dz , at least here.  We probably do elsewhere.
        id = self.id - self._id_offset
        if len(self.Parent) > 0:
            self.dds = self.Parent[0].dds / self.ds.refine_by
        else:
            LE, RE = self.index.grid_left_edge[id, :], self.index.grid_right_edge[id, :]
            self.dds = self.ds.arr((RE - LE) / self.ActiveDimensions, "code_length")
        if self.ds.dimensionality < 2:
            self.dds[1] = 1.0
        if self.ds.dimensionality < 3:
            self.dds[2] = 1.0
        self.field_data["dx"], self.field_data["dy"], self.field_data["dz"] = self.dds


def parse_line(line, grid):
    # grid is a dictionary
    splitup = line.strip().split()
    if chk23("vtk") in splitup:
        grid["vtk_version"] = str23(splitup[-1])
    elif chk23("time=") in splitup:
        time_index = splitup.index(chk23("time="))
        grid["time"] = float(str23(splitup[time_index + 1]).rstrip(","))
        grid["level"] = int(str23(splitup[time_index + 3]).rstrip(","))
        grid["domain"] = int(str23(splitup[time_index + 5]).rstrip(","))
    elif chk23("DIMENSIONS") in splitup:
        grid["dimensions"] = np.array(str23(splitup[-3:])).astype("int")
    elif chk23("ORIGIN") in splitup:
        grid["left_edge"] = np.array(str23(splitup[-3:])).astype("float64")
    elif chk23("SPACING") in splitup:
        grid["dds"] = np.array(str23(splitup[-3:])).astype("float64")
    elif chk23("CELL_DATA") in splitup or chk23("POINT_DATA") in splitup:
        grid["ncells"] = int(str23(splitup[-1]))
    elif chk23("SCALARS") in splitup:
        field = str23(splitup[1])
        grid["read_field"] = field
        grid["read_type"] = "scalar"
    elif chk23("VECTORS") in splitup:
        field = str23(splitup[1])
        grid["read_field"] = field
        grid["read_type"] = "vector"
    elif chk23("time") in splitup:
        time_index = splitup.index(chk23("time"))
        grid["time"] = float(str23(splitup[time_index + 1]))


class AthenaHierarchy(GridIndex):

    grid = AthenaGrid
    _dataset_type = "athena"
    _data_file = None

    def __init__(self, ds, dataset_type="athena"):
        self.dataset = weakref.proxy(ds)
        self.directory = os.path.dirname(self.dataset.filename)
        self.dataset_type = dataset_type
        # for now, the index file is the dataset!
        self.index_filename = os.path.join(os.getcwd(), self.dataset.filename)
        self._fhandle = open(self.index_filename, "rb")
        GridIndex.__init__(self, ds, dataset_type)

        self._fhandle.close()

    def _detect_output_fields(self):
        field_map = {}
        f = open(self.index_filename, "rb")
        line = check_readline(f)
        chkwhile = chk23("")
        while line != chkwhile:
            splitup = line.strip().split()
            chkd = chk23("DIMENSIONS")
            chkc = chk23("CELL_DATA")
            chkp = chk23("POINT_DATA")
            if chkd in splitup:
                field = str23(splitup[-3:])
                grid_dims = np.array(field).astype("int")
                line = check_readline(f)
            elif chkc in splitup or chkp in splitup:
                grid_ncells = int(str23(splitup[-1]))
                line = check_readline(f)
                if np.prod(grid_dims) != grid_ncells:
                    grid_dims -= 1
                    grid_dims[grid_dims == 0] = 1
                if np.prod(grid_dims) != grid_ncells:
                    mylog.error(
                        "product of dimensions %i not equal to number of cells %i",
                        np.prod(grid_dims),
                        grid_ncells,
                    )
                    raise TypeError
                break
            else:
                line = check_readline(f)
        read_table = False
        read_table_offset = f.tell()
        while line != chkwhile:
            splitup = line.strip().split()
            chks = chk23("SCALARS")
            chkv = chk23("VECTORS")
            if chks in line and chks not in splitup:
                splitup = str23(line[line.find(chks) :].strip().split())
            if chkv in line and chkv not in splitup:
                splitup = str23(line[line.find(chkv) :].strip().split())
            if chks in splitup:
                field = ("athena", str23(splitup[1]))
                dtype = str23(splitup[-1]).lower()
                if not read_table:
                    line = check_readline(f)  # Read the lookup table line
                    read_table = True
                field_map[field] = ("scalar", f.tell() - read_table_offset, dtype)
                read_table = False
            elif chkv in splitup:
                field = str23(splitup[1])
                dtype = str23(splitup[-1]).lower()
                for ax in "xyz":
                    field_map[("athena", f"{field}_{ax}")] = (
                        "vector",
                        f.tell() - read_table_offset,
                        dtype,
                    )
            line = check_readline(f)

        f.close()

        self.field_list = list(field_map.keys())
        self._field_map = field_map

    def _count_grids(self):
        self.num_grids = self.dataset.nvtk * self.dataset.nprocs

    def _parse_index(self):
        f = open(self.index_filename, "rb")
        grid = {}
        grid["read_field"] = None
        grid["read_type"] = None
        line = f.readline()
        while grid["read_field"] is None:
            parse_line(line, grid)
            if check_break(line):
                break
            line = f.readline()
        f.close()

        # It seems some datasets have a mismatch between ncells and
        # the actual grid dimensions.
        if np.prod(grid["dimensions"]) != grid["ncells"]:
            grid["dimensions"] -= 1
            grid["dimensions"][grid["dimensions"] == 0] = 1
        if np.prod(grid["dimensions"]) != grid["ncells"]:
            mylog.error(
                "product of dimensions %i not equal to number of cells %i",
                np.prod(grid["dimensions"]),
                grid["ncells"],
            )
            raise TypeError

        # Need to determine how many grids: self.num_grids
        dataset_dir = os.path.dirname(self.index_filename)
        dname = os.path.split(self.index_filename)[-1]
        if dataset_dir.endswith("id0"):
            dname = "id0/" + dname
            dataset_dir = dataset_dir[:-3]

        gridlistread = sglob(
            os.path.join(dataset_dir, f"id*/{dname[4:-9]}-id*{dname[-9:]}")
        )
        gridlistread.insert(0, self.index_filename)
        if "id0" in dname:
            gridlistread += sglob(
                os.path.join(dataset_dir, f"id*/lev*/{dname[4:-9]}*-lev*{dname[-9:]}")
            )
        else:
            gridlistread += sglob(
                os.path.join(dataset_dir, f"lev*/{dname[:-9]}*-lev*{dname[-9:]}")
            )
        ndots = dname.count(".")
        gridlistread = [
            fn for fn in gridlistread if os.path.basename(fn).count(".") == ndots
        ]
        self.num_grids = len(gridlistread)
        dxs = []
        levels = np.zeros(self.num_grids, dtype="int32")
        glis = np.empty((self.num_grids, 3), dtype="float64")
        gdds = np.empty((self.num_grids, 3), dtype="float64")
        gdims = np.ones_like(glis)
        j = 0
        self.grid_filenames = gridlistread
        while j < (self.num_grids):
            f = open(gridlistread[j], "rb")
            gridread = {}
            gridread["read_field"] = None
            gridread["read_type"] = None
            line = f.readline()
            while gridread["read_field"] is None:
                parse_line(line, gridread)
                splitup = line.strip().split()
                if chk23("X_COORDINATES") in splitup:
                    gridread["left_edge"] = np.zeros(3)
                    gridread["dds"] = np.zeros(3)
                    v = np.fromfile(f, dtype=">f8", count=2)
                    gridread["left_edge"][0] = v[0] - 0.5 * (v[1] - v[0])
                    gridread["dds"][0] = v[1] - v[0]
                if chk23("Y_COORDINATES") in splitup:
                    v = np.fromfile(f, dtype=">f8", count=2)
                    gridread["left_edge"][1] = v[0] - 0.5 * (v[1] - v[0])
                    gridread["dds"][1] = v[1] - v[0]
                if chk23("Z_COORDINATES") in splitup:
                    v = np.fromfile(f, dtype=">f8", count=2)
                    gridread["left_edge"][2] = v[0] - 0.5 * (v[1] - v[0])
                    gridread["dds"][2] = v[1] - v[0]
                if check_break(line):
                    break
                line = f.readline()
            f.close()
            levels[j] = gridread.get("level", 0)
            glis[j, 0] = gridread["left_edge"][0]
            glis[j, 1] = gridread["left_edge"][1]
            glis[j, 2] = gridread["left_edge"][2]
            # It seems some datasets have a mismatch between ncells and
            # the actual grid dimensions.
            if np.prod(gridread["dimensions"]) != gridread["ncells"]:
                gridread["dimensions"] -= 1
                gridread["dimensions"][gridread["dimensions"] == 0] = 1
            if np.prod(gridread["dimensions"]) != gridread["ncells"]:
                mylog.error(
                    "product of dimensions %i not equal to number of cells %i",
                    np.prod(gridread["dimensions"]),
                    gridread["ncells"],
                )
                raise TypeError
            gdims[j, 0] = gridread["dimensions"][0]
            gdims[j, 1] = gridread["dimensions"][1]
            gdims[j, 2] = gridread["dimensions"][2]
            # Setting dds=1 for non-active dimensions in 1D/2D datasets
            gridread["dds"][gridread["dimensions"] == 1] = 1.0
            gdds[j, :] = gridread["dds"]

            j = j + 1

        gres = glis + gdims * gdds
        # Now we convert the glis, which were left edges (floats), to indices
        # from the domain left edge.  Then we do a bunch of fixing now that we
        # know the extent of all the grids.
        glis = np.round(
            (glis - self.dataset.domain_left_edge.ndarray_view()) / gdds
        ).astype("int")
        new_dre = np.max(gres, axis=0)
        dre_units = self.dataset.domain_right_edge.uq
        self.dataset.domain_right_edge = np.round(new_dre, decimals=12) * dre_units
        self.dataset.domain_width = (
            self.dataset.domain_right_edge - self.dataset.domain_left_edge
        )
        self.dataset.domain_center = 0.5 * (
            self.dataset.domain_left_edge + self.dataset.domain_right_edge
        )
        self.dataset.domain_dimensions = np.round(
            self.dataset.domain_width / gdds[0]
        ).astype("int")

        if self.dataset.dimensionality <= 2:
            self.dataset.domain_dimensions[2] = 1
        if self.dataset.dimensionality == 1:
            self.dataset.domain_dimensions[1] = 1

        dle = self.dataset.domain_left_edge
        dre = self.dataset.domain_right_edge
        dx_root = (
            self.dataset.domain_right_edge - self.dataset.domain_left_edge
        ) / self.dataset.domain_dimensions

        if self.dataset.nprocs > 1:
            gle_all = []
            gre_all = []
            shapes_all = []
            levels_all = []
            new_gridfilenames = []
            file_offsets = []
            read_dims = []
            for i in range(levels.shape[0]):
                dx = dx_root / self.dataset.refine_by ** (levels[i])
                gle_orig = self.ds.arr(
                    np.round(dle + dx * glis[i], decimals=12), "code_length"
                )
                gre_orig = self.ds.arr(
                    np.round(gle_orig + dx * gdims[i], decimals=12), "code_length"
                )
                bbox = np.array([[le, re] for le, re in zip(gle_orig, gre_orig)])
                psize = get_psize(self.ds.domain_dimensions, self.ds.nprocs)
                gle, gre, shapes, slices = decompose_array(gdims[i], psize, bbox)
                gle_all += gle
                gre_all += gre
                shapes_all += shapes
                levels_all += [levels[i]] * self.dataset.nprocs
                new_gridfilenames += [self.grid_filenames[i]] * self.dataset.nprocs
                file_offsets += [
                    [slc[0].start, slc[1].start, slc[2].start] for slc in slices
                ]
                read_dims += [
                    np.array([gdims[i][0], gdims[i][1], shape[2]], dtype="int64")
                    for shape in shapes
                ]
            self.num_grids *= self.dataset.nprocs
            self.grids = np.empty(self.num_grids, dtype="object")
            self.grid_filenames = new_gridfilenames
            self.grid_left_edge = self.ds.arr(gle_all, "code_length")
            self.grid_right_edge = self.ds.arr(gre_all, "code_length")
            self.grid_dimensions = np.array(
                [shape for shape in shapes_all], dtype="int32"
            )
            gdds = (self.grid_right_edge - self.grid_left_edge) / self.grid_dimensions
            glis = np.round(
                (self.grid_left_edge - self.ds.domain_left_edge) / gdds
            ).astype("int")
            for i in range(self.num_grids):
                self.grids[i] = self.grid(
                    i,
                    self,
                    levels_all[i],
                    glis[i],
                    shapes_all[i],
                    file_offsets[i],
                    read_dims[i],
                )
        else:
            self.grids = np.empty(self.num_grids, dtype="object")
            for i in range(levels.shape[0]):
                self.grids[i] = self.grid(
                    i, self, levels[i], glis[i], gdims[i], [0] * 3, gdims[i]
                )
                dx = dx_root / self.dataset.refine_by ** (levels[i])
                dxs.append(dx)

            dx = self.ds.arr(dxs, "code_length")
            self.grid_left_edge = self.ds.arr(
                np.round(dle + dx * glis, decimals=12), "code_length"
            )
            self.grid_dimensions = gdims.astype("int32")
            self.grid_right_edge = self.ds.arr(
                np.round(self.grid_left_edge + dx * self.grid_dimensions, decimals=12),
                "code_length",
            )
        if self.dataset.dimensionality <= 2:
            self.grid_right_edge[:, 2] = dre[2]
        if self.dataset.dimensionality == 1:
            self.grid_right_edge[:, 1:] = dre[1:]
        self.grid_particle_count = np.zeros([self.num_grids, 1], dtype="int64")

    def _populate_grid_objects(self):
        for g in self.grids:
            g._prepare_grid()
            g._setup_dx()
        self._reconstruct_parent_child()
        self.max_level = self.grid_levels.max()

    def _reconstruct_parent_child(self):
        mask = np.empty(len(self.grids), dtype="int32")
        mylog.debug("First pass; identifying child grids")
        for i, grid in enumerate(self.grids):
            get_box_grids_level(
                self.grid_left_edge[i, :],
                self.grid_right_edge[i, :],
                self.grid_levels[i] + 1,
                self.grid_left_edge,
                self.grid_right_edge,
                self.grid_levels,
                mask,
            )
            grid.Children = [
                g for g in self.grids[mask.astype("bool")] if g.Level == grid.Level + 1
            ]
        mylog.debug("Second pass; identifying parents")
        for grid in self.grids:  # Second pass
            for child in grid.Children:
                child.Parent.append(grid)

    def _get_grid_children(self, grid):
        mask = np.zeros(self.num_grids, dtype="bool")
        grids, grid_ind = self.get_box_grids(grid.LeftEdge, grid.RightEdge)
        mask[grid_ind] = True
        return [g for g in self.grids[mask] if g.Level == grid.Level + 1]

    def _chunk_io(self, dobj, cache=True, local_only=False):
        gobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        for subset in gobjs:
            yield YTDataChunk(
                dobj, "io", [subset], self._count_selection(dobj, [subset]), cache=cache
            )


class AthenaDataset(Dataset):
    _index_class = AthenaHierarchy
    _field_info_class = AthenaFieldInfo
    _dataset_type = "athena"

    def __init__(
        self,
        filename,
        dataset_type="athena",
        storage_filename=None,
        parameters=None,
        units_override=None,
        nprocs=1,
        unit_system="cgs",
        default_species_fields=None,
        magnetic_normalization="gaussian",
    ):
        self.fluid_types += ("athena",)
        self.nprocs = nprocs
        if parameters is None:
            parameters = {}
        self.specified_parameters = parameters.copy()
        if units_override is None:
            units_override = {}
        self._magnetic_factor = get_magnetic_normalization(magnetic_normalization)

        Dataset.__init__(
            self,
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )
        if storage_filename is None:
            storage_filename = self.basename + ".yt"
        self.storage_filename = storage_filename
        # Unfortunately we now have to mandate that the index gets
        # instantiated so that we can make sure we have the correct left
        # and right domain edges.
        self.index

    def _set_code_unit_attributes(self):
        """
        Generates the conversion to various physical _units based on the
        parameter file
        """
        if "length_unit" not in self.units_override:
            self.no_cgs_equiv_length = True
        for unit, cgs in [("length", "cm"), ("time", "s"), ("mass", "g")]:
            # We set these to cgs for now, but they may have been overridden
            if getattr(self, unit + "_unit", None) is not None:
                continue
            mylog.warning("Assuming 1.0 = 1.0 %s", cgs)
            setattr(self, f"{unit}_unit", self.quan(1.0, cgs))
        self.magnetic_unit = np.sqrt(
            self._magnetic_factor
            * self.mass_unit
            / (self.time_unit**2 * self.length_unit)
        )
        self.magnetic_unit.convert_to_units("gauss")
        self.velocity_unit = self.length_unit / self.time_unit

    def _parse_parameter_file(self):
        self._handle = open(self.parameter_filename, "rb")
        # Read the start of a grid to get simulation parameters.
        grid = {}
        grid["read_field"] = None
        line = self._handle.readline()
        while grid["read_field"] is None:
            parse_line(line, grid)
            splitup = line.strip().split()
            if chk23("X_COORDINATES") in splitup:
                grid["left_edge"] = np.zeros(3)
                grid["dds"] = np.zeros(3)
                v = np.fromfile(self._handle, dtype=">f8", count=2)
                grid["left_edge"][0] = v[0] - 0.5 * (v[1] - v[0])
                grid["dds"][0] = v[1] - v[0]
            if chk23("Y_COORDINATES") in splitup:
                v = np.fromfile(self._handle, dtype=">f8", count=2)
                grid["left_edge"][1] = v[0] - 0.5 * (v[1] - v[0])
                grid["dds"][1] = v[1] - v[0]
            if chk23("Z_COORDINATES") in splitup:
                v = np.fromfile(self._handle, dtype=">f8", count=2)
                grid["left_edge"][2] = v[0] - 0.5 * (v[1] - v[0])
                grid["dds"][2] = v[1] - v[0]
            if check_break(line):
                break
            line = self._handle.readline()

        self.domain_left_edge = grid["left_edge"]
        mylog.info(
            "Temporarily setting domain_right_edge = -domain_left_edge. "
            "This will be corrected automatically if it is not the case."
        )
        self.domain_right_edge = -self.domain_left_edge
        self.domain_width = self.domain_right_edge - self.domain_left_edge
        self.domain_dimensions = np.round(self.domain_width / grid["dds"]).astype(
            "int32"
        )
        refine_by = None
        if refine_by is None:
            refine_by = 2
        self.refine_by = refine_by
        dimensionality = 3
        if grid["dimensions"][2] == 1:
            dimensionality = 2
        if grid["dimensions"][1] == 1:
            dimensionality = 1
        if dimensionality <= 2:
            self.domain_dimensions[2] = np.int32(1)
        if dimensionality == 1:
            self.domain_dimensions[1] = np.int32(1)
        if dimensionality != 3 and self.nprocs > 1:
            raise RuntimeError("Virtual grids are only supported for 3D outputs!")
        self.dimensionality = dimensionality
        self.current_time = grid["time"]
        self.cosmological_simulation = False
        self.num_ghost_zones = 0
        self.field_ordering = "fortran"
        self.boundary_conditions = [1] * 6
        self._periodicity = tuple(
            self.specified_parameters.get("periodicity", (True, True, True))
        )
        if "gamma" in self.specified_parameters:
            self.gamma = float(self.specified_parameters["gamma"])
        else:
            self.gamma = 5.0 / 3.0
        dataset_dir = os.path.dirname(self.parameter_filename)
        dname = os.path.split(self.parameter_filename)[-1]
        if dataset_dir.endswith("id0"):
            dname = "id0/" + dname
            dataset_dir = dataset_dir[:-3]

        gridlistread = sglob(
            os.path.join(dataset_dir, f"id*/{dname[4:-9]}-id*{dname[-9:]}")
        )
        if "id0" in dname:
            gridlistread += sglob(
                os.path.join(dataset_dir, f"id*/lev*/{dname[4:-9]}*-lev*{dname[-9:]}")
            )
        else:
            gridlistread += sglob(
                os.path.join(dataset_dir, f"lev*/{dname[:-9]}*-lev*{dname[-9:]}")
            )
        ndots = dname.count(".")
        gridlistread = [
            fn for fn in gridlistread if os.path.basename(fn).count(".") == ndots
        ]
        self.nvtk = len(gridlistread) + 1

        self.current_redshift = 0.0
        self.omega_lambda = 0.0
        self.omega_matter = 0.0
        self.hubble_constant = 0.0
        self.cosmological_simulation = 0
        self.parameters["Time"] = self.current_time  # Hardcode time conversion for now.
        self.parameters[
            "HydroMethod"
        ] = 0  # Hardcode for now until field staggering is supported.
        if "gamma" in self.specified_parameters:
            self.parameters["Gamma"] = self.specified_parameters["gamma"]
        else:
            self.parameters["Gamma"] = 5.0 / 3.0
        self.geometry = self.specified_parameters.get("geometry", "cartesian")
        self._handle.close()
        self.mu = self.specified_parameters.get(
            "mu", compute_mu(self.default_species_fields)
        )

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        if not filename.endswith(".vtk"):
            return False

        with open(filename, "rb") as fh:
            if not re.match(b"# vtk DataFile Version \\d\\.\\d\n", fh.readline(256)):
                return False
            if (
                re.search(
                    b"at time= .*, level= \\d, domain= \\d\n",
                    fh.readline(256),
                )
                is None
            ):
                # vtk Dumps headers start with either "CONSERVED vars" or "PRIMITIVE vars",
                # while vtk output headers start with "Really cool Athena data", but
                # we will consider this an implementation detail and not attempt to
                # match it exactly here.
                # See Athena's user guide for reference
                # https://princetonuniversity.github.io/Athena-Cversion/AthenaDocsUGbtk
                return False
        return True

    @property
    def _skip_cache(self):
        return True

    def __str__(self):
        return self.basename.rsplit(".", 1)[0]



import sys
import weakref
from numbers import Real
from typing import Any, Dict, List, Optional, Tuple, Type, Union

import numpy as np
import unyt as un
from matplotlib.colors import Colormap, LogNorm, Normalize, SymLogNorm
from packaging.version import Version
from unyt import unyt_quantity

from yt._typing import Quantity, Unit
from yt.config import ytcfg
from yt.funcs import get_brewer_cmap, is_sequence, mylog
from yt.visualization._commons import MPL_VERSION
from yt.visualization.color_maps import _get_cmap

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class NormHandler:
    """
    A bookkeeper class that can hold a fully defined norm object, or dynamically
    build one on demand according to a set of constraints.

    If a fully defined norm object is added, any existing constraints are
    dropped, and vice versa. These rules are implemented with properties and
    watcher patterns.

    It also keeps track of display units so that vmin, vmax and linthresh can be
    updated with implicit units.
    """

    # using slots here to minimize the risk of introducing bugs
    # since attributes names are essential to this class's implementation
    __slots__ = (
        "data_source",
        "ds",
        "_display_units",
        "_vmin",
        "_vmax",
        "_dynamic_range",
        "_norm_type",
        "_linthresh",
        "_norm_type",
        "_norm",
    )
    _constraint_attrs: List[str] = [
        "vmin",
        "vmax",
        "dynamic_range",
        "norm_type",
        "linthresh",
    ]

    def __init__(
        self,
        data_source,
        *,
        display_units: un.Unit,
        vmin: Optional[un.unyt_quantity] = None,
        vmax: Optional[un.unyt_quantity] = None,
        dynamic_range: Optional[float] = None,
        norm_type: Optional[Type[Normalize]] = None,
        norm: Optional[Normalize] = None,
        linthresh: Optional[float] = None,
    ):
        self.data_source = weakref.proxy(data_source)
        self.ds = data_source.ds  # should already be a weakref proxy
        self._display_units = display_units

        self._norm = norm
        self._vmin = vmin
        self._vmax = vmax
        self._dynamic_range = dynamic_range
        self._norm_type = norm_type
        self._linthresh = linthresh

        if self.has_norm and self.has_constraints:
            raise TypeError(
                "NormHandler input is malformed. "
                "A norm cannot be passed along other constraints."
            )

    def _get_constraints(self) -> Dict[str, Any]:
        return {
            attr: getattr(self, attr)
            for attr in self.__class__._constraint_attrs
            if getattr(self, attr) is not None
        }

    @property
    def has_constraints(self) -> bool:
        return bool(self._get_constraints())

    def _reset_constraints(self) -> None:
        constraints = self._get_constraints()
        if not constraints:
            return

        msg = ", ".join([f"{name}={value}" for name, value in constraints.items()])
        mylog.warning("Dropping norm constraints (%s)", msg)
        for name in constraints.keys():
            setattr(self, name, None)

    @property
    def has_norm(self) -> bool:
        return self._norm is not None

    def _reset_norm(self):
        if not self.has_norm:
            return
        mylog.warning("Dropping norm (%s)", self.norm)
        self._norm = None

    def to_float(self, val: un.unyt_quantity) -> float:
        return float(val.to(self.display_units).d)

    def to_quan(self, val) -> un.unyt_quantity:
        if isinstance(val, un.unyt_quantity):
            return self.ds.quan(val)
        elif (
            is_sequence(val)
            and len(val) == 2
            and isinstance(val[0], Real)
            and isinstance(val[1], (str, un.Unit))
        ):
            return self.ds.quan(*val)
        elif isinstance(val, Real):
            return self.ds.quan(val, self.display_units)
        else:
            raise TypeError(f"Could not convert {val!r} to unyt_quantity")

    @property
    def display_units(self) -> un.Unit:
        return self._display_units

    @display_units.setter
    def display_units(self, newval: Unit) -> None:
        self._display_units = un.Unit(newval, registry=self.ds.unit_registry)

    def _set_quan_attr(
        self, attr: str, newval: Optional[Union[Quantity, float]]
    ) -> None:
        if newval is None:
            setattr(self, attr, None)
        else:
            try:
                quan = self.to_quan(newval)
            except TypeError as exc:
                raise TypeError(
                    "Expected None, a float, or a unyt_quantity, "
                    f"received {newval} with type {type(newval)}"
                ) from exc
            else:
                setattr(self, attr, quan)

    @property
    def vmin(self) -> Optional[Union[un.unyt_quantity, Literal["min"]]]:
        return self._vmin

    @vmin.setter
    def vmin(self, newval: Optional[Union[Quantity, float, Literal["min"]]]) -> None:
        self._reset_norm()
        if newval == "min":
            self._vmin = "min"
        else:
            self._set_quan_attr("_vmin", newval)

    @property
    def vmax(self) -> Optional[Union[un.unyt_quantity, Literal["max"]]]:
        return self._vmax

    @vmax.setter
    def vmax(self, newval: Optional[Union[Quantity, float, Literal["max"]]]) -> None:
        self._reset_norm()
        if newval == "max":
            self._vmax = "max"
        else:
            self._set_quan_attr("_vmax", newval)

    @property
    def dynamic_range(self) -> Optional[float]:
        return self._dynamic_range

    @dynamic_range.setter
    def dynamic_range(self, newval: Optional[float]) -> None:
        if newval is None:
            return

        try:
            newval = float(newval)
        except TypeError:
            raise TypeError(
                f"Expected a float. Received {newval} with type {type(newval)}"
            ) from None

        if newval == 0:
            raise ValueError("Dynamic range cannot be zero.")

        if newval == 1:
            raise ValueError("Dynamic range cannot be unity.")

        self._reset_norm()
        self._dynamic_range = newval

    def get_dynamic_range(
        self, dvmin: Optional[float], dvmax: Optional[float]
    ) -> Tuple[float, float]:
        if self.dynamic_range is None:
            raise RuntimeError(
                "Something went terribly wrong in setting up a dynamic range"
            )

        if self.vmax is None:
            if self.vmin is None:
                raise TypeError(
                    "Cannot set dynamic range with neither "
                    "vmin and vmax being constrained."
                )
            if dvmin is None:
                raise RuntimeError(
                    "Something went terribly wrong in setting up a dynamic range"
                )
            return dvmin, dvmin * self.dynamic_range
        elif self.vmin is None:
            if dvmax is None:
                raise RuntimeError(
                    "Something went terribly wrong in setting up a dynamic range"
                )
            return dvmax / self.dynamic_range, dvmax
        else:
            raise TypeError(
                "Cannot set dynamic range with both "
                "vmin and vmax already constrained."
            )

    @property
    def norm_type(self) -> Optional[Type[Normalize]]:
        return self._norm_type

    @norm_type.setter
    def norm_type(self, newval: Optional[Type[Normalize]]) -> None:
        if not (
            newval is None
            or (isinstance(newval, type) and issubclass(newval, Normalize))
        ):
            raise TypeError(
                "Expected a subclass of matplotlib.colors.Normalize, "
                f"received {newval} with type {type(newval)}"
            )
        self._reset_norm()
        if newval is not SymLogNorm:
            self.linthresh = None
        self._norm_type = newval

    @property
    def norm(self) -> Optional[Normalize]:
        return self._norm

    @norm.setter
    def norm(self, newval: Normalize) -> None:
        if not isinstance(newval, Normalize):
            raise TypeError(
                "Expected a matplotlib.colors.Normalize object, "
                f"received {newval} with type {type(newval)}"
            )
        self._reset_constraints()
        self._norm = newval

    @property
    def linthresh(self) -> Optional[float]:
        return self._linthresh

    @linthresh.setter
    def linthresh(self, newval: Optional[Union[Quantity, float]]) -> None:
        self._reset_norm()
        self._set_quan_attr("_linthresh", newval)
        if self._linthresh is not None and self._linthresh <= 0:
            raise ValueError(
                f"linthresh can only be set to strictly positive values, got {newval}"
            )
        if newval is not None:
            self.norm_type = SymLogNorm

    def get_norm(self, data: np.ndarray, *args, **kw) -> Normalize:
        if self.has_norm:
            return self.norm

        dvmin = dvmax = None

        finite_values_mask = np.isfinite(data)
        # FUTURE: when the minimal supported version of numpy reaches 1.16 or newer,
        # this complicated conditional can be simplified into
        # if self.vmin not in (None, "min"):
        if self.vmin is not None and not (
            isinstance(self.vmin, str) and self.vmin == "min"
        ):
            dvmin = self.to_float(self.vmin)
        elif np.any(finite_values_mask):
            dvmin = self.to_float(np.nanmin(data[finite_values_mask]))

        # FUTURE: see above
        if self.vmax is not None and not (
            isinstance(self.vmax, str) and self.vmax == "max"
        ):
            dvmax = self.to_float(self.vmax)
        elif np.any(finite_values_mask):
            dvmax = self.to_float(np.nanmax(data[finite_values_mask]))

        if self.dynamic_range is not None:
            dvmin, dvmax = self.get_dynamic_range(dvmin, dvmax)

        if dvmin is None:
            dvmin = 1 * getattr(data, "units", 1)
        kw.setdefault("vmin", dvmin)

        if dvmax is None:
            dvmax = 1 * getattr(data, "units", 1)
        kw.setdefault("vmax", dvmax)

        if data.ndim == 3:
            assert data.shape[-1] == 4
            # this is an RGBA array, only linear normalization makes sense here
            norm_type = Normalize
        elif self.norm_type is not None:
            # this is a convenience mechanism for backward compat,
            # allowing to toggle between lin and log scaling without detailed user input
            norm_type = self.norm_type
        else:
            if kw["vmin"] == kw["vmax"] or not np.any(finite_values_mask):
                norm_type = Normalize
            elif kw["vmin"] <= 0:
                # note: see issue 3944 (and PRs and issues linked therein) for a
                # discussion on when to switch to SymLog and related questions
                # of how to calculate a default linthresh value.
                norm_type = SymLogNorm
            else:
                norm_type = LogNorm

        if norm_type is SymLogNorm:
            if self.linthresh is not None:
                linthresh = self.to_float(self.linthresh)
            else:
                linthresh = self._guess_linthresh(data[finite_values_mask])

            kw.setdefault("linthresh", linthresh)
            if MPL_VERSION >= Version("3.2"):
                # note that this creates an inconsistency between mpl versions
                # since the default value previous to mpl 3.4.0 is np.e
                # but it is only exposed since 3.2.0
                kw.setdefault("base", 10)

        return norm_type(*args, **kw)

    def _guess_linthresh(self, finite_plot_data):
        # finite_plot_data is the ImageArray or ColorbarHandler data, already
        # filtered to be finite values

        # get the extrema for the negative and positive values separately
        # neg_min -> neg_max -> 0 -> pos_min -> pos_max
        def get_minmax(data):
            if len(data) > 0:
                return np.nanmin(data), np.nanmax(data)
            return None, None

        pos_min, pos_max = get_minmax(finite_plot_data[finite_plot_data > 0])
        neg_min, neg_max = get_minmax(finite_plot_data[finite_plot_data < 0])

        has_pos = pos_min is not None
        has_neg = neg_min is not None

        # the starting guess is the absolute value of the point closest to 0
        # (remember: neg_max is closer to 0 than neg_min)
        if has_pos and has_neg:
            linthresh = np.min((-neg_max, pos_min))
        elif has_pos:
            linthresh = pos_min
        elif has_neg:
            linthresh = -neg_max
        else:
            # this condition should be handled before here in get_norm
            raise RuntimeError("No finite data points.")

        log10_linthresh = np.log10(linthresh)

        # if either the pos or neg ranges exceed cutoff_sigdigs, then
        # linthresh is shifted to decrease the range to avoid floating point
        # precision errors in the normalization.
        cutoff_sigdigs = 15  # max allowable range in significant digits
        if has_pos and np.log10(pos_max) - log10_linthresh > cutoff_sigdigs:
            linthresh = pos_max / (10.0**cutoff_sigdigs)
            log10_linthresh = np.log10(linthresh)

        if has_neg and np.log10(-neg_min) - log10_linthresh > cutoff_sigdigs:
            linthresh = np.abs(neg_min) / (10.0**cutoff_sigdigs)

        if isinstance(linthresh, unyt_quantity):
            # if the original plot_data has units, linthresh will have units here
            return self.to_float(linthresh)

        return linthresh


class ColorbarHandler:
    __slots__ = ("_draw_cbar", "_draw_minorticks", "_cmap", "_background_color")

    def __init__(
        self,
        *,
        draw_cbar: bool = True,
        draw_minorticks: bool = True,
        cmap: Optional[Union[Colormap, str]] = None,
        background_color: Optional[str] = None,
    ):
        self._draw_cbar = draw_cbar
        self._draw_minorticks = draw_minorticks
        self._cmap: Optional[Colormap] = None
        self.cmap = cmap
        self._background_color = background_color

    @property
    def draw_cbar(self) -> bool:
        return self._draw_cbar

    @draw_cbar.setter
    def draw_cbar(self, newval) -> None:
        if not isinstance(newval, bool):
            raise TypeError(
                f"Excpected a boolean, got {newval} with type {type(newval)}"
            )
        self._draw_cbar = newval

    @property
    def draw_minorticks(self) -> bool:
        return self._draw_minorticks

    @draw_minorticks.setter
    def draw_minorticks(self, newval) -> None:
        if not isinstance(newval, bool):
            raise TypeError(
                f"Excpected a boolean, got {newval} with type {type(newval)}"
            )
        self._draw_minorticks = newval

    @property
    def cmap(self) -> Colormap:
        return self._cmap or _get_cmap(ytcfg.get("yt", "default_colormap"))

    @cmap.setter
    def cmap(self, newval) -> None:
        if isinstance(newval, Colormap) or newval is None:
            self._cmap = newval
        elif isinstance(newval, str):
            self._cmap = _get_cmap(newval)
        elif is_sequence(newval):
            # tuple colormaps are from palettable (or brewer2mpl)
            self._cmap = get_brewer_cmap(newval)
        else:
            raise TypeError(
                "Expected a colormap object or name, "
                f"got {newval} with type {type(newval)}"
            )

    @property
    def background_color(self) -> Any:
        return self._background_color or "white"

    @background_color.setter
    def background_color(self, newval: Any):
        # not attempting to constrain types here because
        # down the line it really depends on matplotlib.axes.Axes.set_faceolor
        # which is very type-flexibile
        if newval is None:
            self._background_color = self.cmap(0)
        else:
            self._background_color = newval

    @property
    def has_background_color(self) -> bool:
        return self._background_color is not None



import numpy as np

from yt.loaders import load
from yt.testing import (
    assert_almost_equal,
    assert_equal,
    fake_amr_ds,
    fake_random_ds,
    requires_file,
)


def setup():
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def test_cut_region():
    # We decompose in different ways
    for nprocs in [1, 2, 4, 8]:
        ds = fake_random_ds(
            64,
            nprocs=nprocs,
            fields=("density", "temperature", "velocity_x"),
            units=("g/cm**3", "K", "cm/s"),
        )
        # We'll test two objects
        dd = ds.all_data()
        r = dd.cut_region(
            [
                "obj[('gas', 'temperature')] > 0.5",
                "obj[('gas', 'density')] < 0.75",
                "obj[('gas', 'velocity_x')] > 0.25",
            ]
        )
        t = (
            (dd[("gas", "temperature")] > 0.5)
            & (dd[("gas", "density")] < 0.75)
            & (dd[("gas", "velocity_x")] > 0.25)
        )
        assert_equal(np.all(r[("gas", "temperature")] > 0.5), True)
        assert_equal(np.all(r[("gas", "density")] < 0.75), True)
        assert_equal(np.all(r[("gas", "velocity_x")] > 0.25), True)
        assert_equal(np.sort(dd[("gas", "density")][t]), np.sort(r[("gas", "density")]))
        assert_equal(np.sort(dd[("index", "x")][t]), np.sort(r[("index", "x")]))
        r2 = r.cut_region(["obj[('gas', 'temperature')] < 0.75"])
        t2 = r[("gas", "temperature")] < 0.75
        assert_equal(
            np.sort(r2[("gas", "temperature")]), np.sort(r[("gas", "temperature")][t2])
        )
        assert_equal(np.all(r2[("gas", "temperature")] < 0.75), True)

        # Now we can test some projections
        dd = ds.all_data()
        cr = dd.cut_region(["obj[('index', 'ones')] > 0"])
        for weight in [None, ("gas", "density")]:
            p1 = ds.proj(("gas", "density"), 0, data_source=dd, weight_field=weight)
            p2 = ds.proj(("gas", "density"), 0, data_source=cr, weight_field=weight)
            for f in p1.field_data:
                assert_almost_equal(p1[f], p2[f])
        cr = dd.cut_region(["obj[('gas', 'density')] > 0.25"])
        p2 = ds.proj(("gas", "density"), 2, data_source=cr)
        assert_equal(p2[("gas", "density")].max() > 0.25, True)
        p2 = ds.proj(
            ("gas", "density"), 2, data_source=cr, weight_field=("gas", "density")
        )
        assert_equal(p2[("gas", "density")].max() > 0.25, True)


def test_region_and_particles():
    ds = fake_amr_ds(particles=10000)

    ad = ds.all_data()
    reg = ad.cut_region('obj[("index", "x")] < .5')

    mask = ad[("all", "particle_position_x")] < 0.5
    expected = np.sort(ad[("all", "particle_position_x")][mask].value)
    result = np.sort(reg[("all", "particle_position_x")])

    assert_equal(expected.shape, result.shape)
    assert_equal(expected, result)


ISOGAL = "IsolatedGalaxy/galaxy0030/galaxy0030"


@requires_file(ISOGAL)
def test_region_chunked_read():
    # see #2104
    ds = load("IsolatedGalaxy/galaxy0030/galaxy0030")

    sp = ds.sphere((0.5, 0.5, 0.5), (2, "kpc"))
    dense_sp = sp.cut_region(['obj[("gas", "H_p0_number_density")]>= 1e-2'])
    dense_sp.quantities.angular_momentum_vector()


@requires_file(ISOGAL)
def test_chained_cut_region():
    # see Issue #2233
    ds = load(ISOGAL)
    base = ds.disk([0.5, 0.5, 0.5], [0, 0, 1], (4, "kpc"), (10, "kpc"))
    c1 = "(obj[('index', 'cylindrical_radius')].in_units('kpc') > 2.0)"
    c2 = "(obj[('gas', 'density')].to('g/cm**3') > 1e-26)"

    cr12 = base.cut_region([c1, c2])
    cr1 = base.cut_region([c1])
    cr12c = cr1.cut_region([c2])

    field = ("index", "cell_volume")
    assert_equal(
        cr12.quantities.total_quantity(field), cr12c.quantities.total_quantity(field)
    )



import numpy as np

import yt

OCT_MASK_LIST = [
    8,
    0,
    0,
    0,
    0,
    8,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    8,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
]


def test_octree():
    # See Issue #1272
    octree_mask = np.array(OCT_MASK_LIST, dtype=np.uint8)

    quantities = {}
    quantities[("gas", "density")] = np.ones((22, 1), dtype="float64")

    bbox = np.array([[-10.0, 10.0], [-10.0, 10.0], [-10.0, 10.0]])

    ds = yt.load_octree(
        octree_mask=octree_mask,
        data=quantities,
        bbox=bbox,
        num_zones=1,
        partial_coverage=0,
    )

    proj = ds.proj(("gas", "density"), "x")
    proj[("gas", "density")]



"""
Particle Deposition onto Octs




"""

cimport cython
cimport numpy as np
from libc.float cimport DBL_MANT_DIG
from libc.math cimport frexp, ldexp, sqrt

DEF ORDER_MAX=20
DEF INDEX_MAX_64=2097151
# TODO: Handle error for indices past max
DEF XSHIFT=2
DEF YSHIFT=1
DEF ZSHIFT=0

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline np.int64_t ifrexp(np.float64_t x, np.int64_t *e):
    cdef np.float64_t m
    cdef int e0 = 0
    m = frexp(x,&e0)
    e[0] = <np.int64_t>e0
    return <np.int64_t>ldexp(m,<int>DBL_MANT_DIG)

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline np.int64_t msdb(np.int64_t a, np.int64_t b):
    """Get the most significant differing bit between a and b."""
    cdef np.int64_t c, ndx
    c = a ^ b
    ndx = 0
    while (0 < c):
        c = (c >> 1)
        ndx+=1
    return ndx

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline np.int64_t xor_msb(np.float64_t a, np.float64_t b):
    """Get the exponent of the highest differing bit between a and b"""
    # Get mantissa and exponents for each number
    cdef np.int64_t a_m, a_e, b_m, b_e, x, y, z
    b_e = 0
    a_e = 0
    a_m = ifrexp(a,&a_e)
    b_m = ifrexp(b,&b_e)
    x = <np.int64_t> ((a_e+1)*DBL_MANT_DIG)
    y = <np.int64_t> ((b_e+1)*DBL_MANT_DIG)
    # Compare mantissa if exponents equal
    if x == y:
        if a_m == b_m: return 0
        z = msdb(a_m,b_m)
        #if 1: return z
        x = x - z
        return x-1 # required so that xor_msb(0.0,1.0)!=xor_msb(1.0,1.0)
    # Otherwise return largest exponent
    if y < x:
        return x
    else:
        return y

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline int compare_floats_morton(np.float64_t p[3], np.float64_t q[3]):
    cdef int j, out, dim
    cdef np.int64_t x, y
    x = -9999999999
    y = 0
    dim = 0
    for j in range(3):#[::-1]:
        y = xor_msb(p[j],q[j])
        if x < y:
           x = y
           dim = j
    if p[dim] < q[dim]:
        out = 1
    else:
        out = 0
    return out

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline np.float64_t euclidean_distance(np.float64_t[:] p, np.float64_t[:] q):
    cdef int j
    cdef np.float64_t d
    d = 0.0
    for j in range(3):
        d+=(p[j]-q[j])**2
    return sqrt(d)

# Todo: allow radius reported independently in each dimension for rectangular domain
@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline np.float64_t smallest_quadtree_box(np.float64_t p[3], np.float64_t q[3], np.int32_t order,
                                               np.float64_t DLE[3], np.float64_t DRE[3],
                                               np.float64_t *cx, np.float64_t *cy, np.float64_t *cz):
    cdef int j
    cdef np.float64_t c[3]
    cdef np.uint64_t pidx[3]
    # cdef np.uint64_t qidx[3]
    for j in range(3):
        pidx[j] = 0
        # qidx[j] = 0
    cdef np.uint64_t pidx_next[3]
    cdef np.uint64_t qidx_next[3]
    cdef np.float64_t dds[3]
    cdef np.float64_t rad
    cdef int lvl = 0
    cdef int done = 0
    while not done:
        if (lvl+1 >= order):
            done = 1
        for j in range(3):
            dds[j] = (DRE[j] - DLE[j])/(1 << (<int> lvl+1))
            pidx_next[j] = <np.uint64_t>((p[j] - DLE[j])/dds[j])
            qidx_next[j] = <np.uint64_t>((q[j] - DLE[j])/dds[j])
        for j in range(3):
            if pidx_next[j]!=qidx_next[j]:
                done = 1
                break
        if not done:
            for j in range(3):
                pidx[j] = pidx_next[j]
                # qidx[j] = qidx_next[j]
            lvl+=1
    rad = 0.0
    for j in range(3):
        dds[j] = (DRE[j] - DLE[j])/(1 << lvl)
        c[j] = dds[j]*(<np.float64_t>pidx[j]+0.5)
        rad+=((dds[j]/2.0)**2)
    cx[0] = c[0]
    cy[0] = c[1]
    cz[0] = c[2]
    return sqrt(rad)

#-----------------------------------------------------------------------------
# 21 bits spread over 64 with 3 bits in between
@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline np.uint64_t spread_64bits_by3(np.uint64_t x):
    x=(x&(<np.uint64_t>0x00000000001FFFFF))
    x=(x|(x<<20))*(<np.uint64_t>0x000001FFC00003FF)

#-----------------------------------------------------------------------------
# 21 bits spread over 64 with 2 bits in between
@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline np.uint64_t spread_64bits_by2(np.uint64_t x):
    # This magic comes from http://stackoverflow.com/questions/1024754/how-to-compute-a-3d-morton-number-interleave-the-bits-of-3-ints
    # Only reversible up to 2097151
    # Select highest 21 bits (Required to be reversible to 21st bit)
    # x = ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---k jihg fedc ba98 7654 3210
    x=(x&(<np.uint64_t>0x00000000001FFFFF))
    # x = ---- ---- ---- ---- ---- ---k jihg fedc ba-- ---- ---- ---- ---- --98 7654 3210
    x=(x|(x<<20))&(<np.uint64_t>0x000001FFC00003FF)
    # x = ---- ---- ---- -kji hgf- ---- ---- -edc ba-- ---- ---- 9876 5--- ---- ---4 3210
    x=(x|(x<<10))&(<np.uint64_t>0x0007E007C00F801F)
    # x = ---- ---- -kji h--- -gf- ---- -edc ---- ba-- ---- 987- ---6 5--- ---4 32-- --10
    x=(x|(x<<4))&(<np.uint64_t>0x00786070C0E181C3)
    # x = ---- ---k ji-- h--g --f- ---e d--c --b- -a-- --98 --7- -6-- 5--- -43- -2-- 1--0
    x=(x|(x<<2))&(<np.uint64_t>0x0199219243248649)
    # x = ---- -kj- -i-- h--g --f- -e-- d--c --b- -a-- 9--8 --7- -6-- 5--4 --3- -2-- 1--0
    x=(x|(x<<2))&(<np.uint64_t>0x0649249249249249)
    # x = ---k --j- -i-- h--g --f- -e-- d--c --b- -a-- 9--8 --7- -6-- 5--4 --3- -2-- 1--0
    x=(x|(x<<2))&(<np.uint64_t>0x1249249249249249)
    return x

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline np.uint64_t compact_64bits_by2(np.uint64_t x):
    # Reversed magic
    x=x&(<np.uint64_t>0x1249249249249249)
    x=(x|(x>>2))&(<np.uint64_t>0x0649249249249249)
    x=(x|(x>>2))&(<np.uint64_t>0x0199219243248649)
    x=(x|(x>>2))&(<np.uint64_t>0x00786070C0E181C3)
    x=(x|(x>>4))&(<np.uint64_t>0x0007E007C00F801F)
    x=(x|(x>>10))&(<np.uint64_t>0x000001FFC00003FF)
    x=(x|(x>>20))&(<np.uint64_t>0x00000000001FFFFF)
    return x

#-----------------------------------------------------------------------------
# 10 bits spread over 32 with 2 bits in between
@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline np.uint32_t spread_32bits_by2(np.uint32_t x):
    # Only reversible up to 1023
    # Select highest 10 bits (Required to be reversible to 10st bit)
    # x = ---- ---- ---- ---- ---- --98 7654 3210
    x=(x&(<np.uint32_t>0x000003FF))
    # x = ---- --98 ---- ---- ---- ---- 7654 3210
    x=(x|(x<<16))&(<np.uint32_t>0xFF0000FF)
    # x = ---- --98 ---- ---- 7654 ---- ---- 3210
    x=(x|(x<<8))&(<np.uint32_t>0x0300F00F)
    # x = ---- --98 ---- 76-- --54 ---- 32-- --10
    x=(x|(x<<4))&(<np.uint32_t>0x030C30C3)
    # x = ---- 9--8 --7- -6-- 5--4 --3- -2-- 1--0
    x=(x|(x<<2))&(<np.uint32_t>0x09249249)
    return x

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline np.uint32_t compact_32bits_by2(np.uint32_t x):
    # Reversed magic
    x=x&(<np.uint32_t>0x09249249)
    x=(x|(x>>2))&(<np.uint32_t>0x030C30C3)
    x=(x|(x>>4))&(<np.uint32_t>0x0300F00F)
    x=(x|(x>>8))&(<np.uint32_t>0xFF0000FF)
    x=(x|(x>>16))&(<np.uint32_t>0x000003FF)
    return x

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline np.uint64_t masked_merge_64bit(np.uint64_t a, np.uint64_t b, np.uint64_t mask):
    # https://graphics.stanford.edu/~seander/bithacks.html#MaskedMerge
    return a ^ ((a ^ b) & mask)

@cython.cdivision(True)
cdef inline np.uint64_t encode_morton_64bit(np.uint64_t x_ind, np.uint64_t y_ind, np.uint64_t z_ind):
    cdef np.uint64_t mi
    mi = 0
    mi |= spread_64bits_by2(z_ind)<<ZSHIFT
    mi |= spread_64bits_by2(y_ind)<<YSHIFT
    mi |= spread_64bits_by2(x_ind)<<XSHIFT
    return mi

@cython.cdivision(True)
cdef inline void decode_morton_64bit(np.uint64_t mi, np.uint64_t *p):
    p[0] = compact_64bits_by2(mi>>XSHIFT)
    p[1] = compact_64bits_by2(mi>>YSHIFT)
    p[2] = compact_64bits_by2(mi>>ZSHIFT)

@cython.cdivision(True)
cdef inline np.uint64_t bounded_morton(np.float64_t x, np.float64_t y, np.float64_t z,
                               np.float64_t *DLE, np.float64_t *DRE, np.int32_t order):
    cdef int i
    cdef np.float64_t dds[3]
    cdef np.uint64_t x_ind, y_ind, z_ind
    cdef np.uint64_t mi
    for i in range(3):
        dds[i] = (DRE[i] - DLE[i]) / (1 << order)
    x_ind = <np.uint64_t> ((x - DLE[0])/dds[0])
    y_ind = <np.uint64_t> ((y - DLE[1])/dds[1])
    z_ind = <np.uint64_t> ((z - DLE[2])/dds[2])
    mi = encode_morton_64bit(x_ind,y_ind,z_ind)
    return mi

@cython.cdivision(True)
cdef inline np.uint64_t bounded_morton_relative(np.float64_t x, np.float64_t y, np.float64_t z,
                               np.float64_t *DLE, np.float64_t *DRE,
                               np.int32_t order1, np.int32_t order2):
    cdef int i
    cdef np.float64_t dds1[3]
    cdef np.float64_t dds2[3]
    cdef np.float64_t DLE2[3]
    cdef np.uint64_t x_ind, y_ind, z_ind
    cdef np.uint64_t mi2
    for i in range(3):
        dds1[i] = (DRE[i] - DLE[i]) / (1 << order1)
        dds2[i] = dds1[i] / (1 << order2)
    DLE2[0] = <np.float64_t> (<np.uint64_t> ((x - DLE[0])/dds1[0])) * dds1[0]
    DLE2[1] = <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]
    DLE2[2] = <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
    x_ind = <np.uint64_t> ((x - DLE2[0])/dds2[0])
    y_ind = <np.uint64_t> ((y - DLE2[1])/dds2[1])
    z_ind = <np.uint64_t> ((z - DLE2[2])/dds2[2])
    mi2 = encode_morton_64bit(x_ind,y_ind,z_ind)
    return mi2


# This doesn't seem to be much, if at all, faster...
@cython.cdivision(True)
cdef inline np.uint64_t bounded_morton_dds(np.float64_t x, np.float64_t y, np.float64_t z,
                               np.float64_t *DLE, np.float64_t *dds):
    cdef np.uint64_t x_ind, y_ind, z_ind
    cdef np.uint64_t mi
    x_ind = <np.uint64_t> ((x - DLE[0])/dds[0])
    y_ind = <np.uint64_t> ((y - DLE[1])/dds[1])
    z_ind = <np.uint64_t> ((z - DLE[2])/dds[2])
    mi = encode_morton_64bit(x_ind,y_ind,z_ind)
    return mi

@cython.cdivision(True)
cdef inline np.uint64_t bounded_morton_relative_dds(np.float64_t x, np.float64_t y, np.float64_t z,
                               np.float64_t *DLE, np.float64_t *dds1, np.float64_t *dds2):
    cdef np.float64_t DLE2[3]
    cdef np.uint64_t x_ind, y_ind, z_ind
    cdef np.uint64_t mi2
    DLE2[0] = <np.float64_t> (<np.uint64_t> ((x - DLE[0])/dds1[0])) * dds1[0]
    DLE2[1] = <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]
    DLE2[2] = <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
    x_ind = <np.uint64_t> ((x - DLE2[0])/dds2[0])
    y_ind = <np.uint64_t> ((y - DLE2[1])/dds2[1])
    z_ind = <np.uint64_t> ((z - DLE2[2])/dds2[2])
    mi2 = encode_morton_64bit(x_ind,y_ind,z_ind)
    return mi2


@cython.cdivision(True)
cdef inline np.uint64_t bounded_morton_split_dds(np.float64_t x, np.float64_t y, np.float64_t z,
                               np.float64_t *DLE, np.float64_t *dds, np.uint64_t *p):
    cdef np.uint64_t mi
    p[0] = <np.uint64_t> ((x - DLE[0])/dds[0])
    p[1] = <np.uint64_t> ((y - DLE[1])/dds[1])
    p[2] = <np.uint64_t> ((z - DLE[2])/dds[2])
    mi = encode_morton_64bit(p[0], p[1], p[2])
    return mi

@cython.cdivision(True)
cdef inline np.uint64_t bounded_morton_split_relative_dds(np.float64_t x, np.float64_t y, np.float64_t z,
                               np.float64_t *DLE, np.float64_t *dds1, np.float64_t *dds2,
                               np.uint64_t *p2):
    cdef np.float64_t DLE2[3]
    cdef np.uint64_t mi2
    DLE2[0] = DLE[0] + <np.float64_t> (<np.uint64_t> ((x - DLE[0])/dds1[0])) * dds1[0]
    DLE2[1] = DLE[1] + <np.float64_t> (<np.uint64_t> ((y - DLE[1])/dds1[1])) * dds1[1]
    DLE2[2] = DLE[2] + <np.float64_t> (<np.uint64_t> ((z - DLE[2])/dds1[2])) * dds1[2]
    p2[0] = <np.uint64_t> ((x - DLE2[0])/dds2[0])
    p2[1] = <np.uint64_t> ((y - DLE2[1])/dds2[1])
    p2[2] = <np.uint64_t> ((z - DLE2[2])/dds2[2])
    mi2 = encode_morton_64bit(p2[0], p2[1], p2[2])
    return mi2


cdef np.uint32_t morton_neighbors_coarse(np.uint64_t mi1, np.uint64_t max_index1,
                                         bint periodicity[3], np.uint32_t nn,
                                         np.uint32_t[:,:] index,
                                         np.uint64_t[:,:] ind1_n,
                                         np.uint64_t[:] neighbors)

cdef np.uint32_t morton_neighbors_refined(np.uint64_t mi1, np.uint64_t mi2,
                                          np.uint64_t max_index1, np.uint64_t max_index2,
                                          bint periodicity[3], np.uint32_t nn,
                                          np.uint32_t[:,:] index,
                                          np.uint64_t[:,:] ind1_n,
                                          np.uint64_t[:,:] ind2_n,
                                          np.uint64_t[:] neighbors1,
                                          np.uint64_t[:] neighbors2)
